from typing import Iterable
import numpy as np
from numpy.typing import NDArray

class Matrix:
    def __init__(self, coefs, inv_coefs):
        self.coefs = coefs
        self.inv_coefs = inv_coefs

    def inverse(self):
        return Matrix(self.inv_coefs, self.coefs)
    def transform_vect(self, vect):
        return (self * Matrix.vector(vect)).flatten()
    def transform_point(self, vect: Iterable[float]) -> NDArray[float]:
        return (self * Matrix.vector((*vect, 1.0))).flatten_homogeneous()
    def flatten(self):
        return self.coefs.flatten()
    def flatten_homogeneous(self):
        ret = self.coefs.flatten()
        dim = len(ret)
        return np.delete(ret, dim-1) / ret[dim-1]
    @property
    def shape(self):
        return self.coefs.shape

    @property
    def shape(self):
        return self.coefs.shape

    @staticmethod
    def extend(n, i):
        ident = np.fromfunction(lambda i,j: np.where(i==j, 1.0, 0.0), (n, n))
        return Matrix(np.delete(ident, i, axis=1), np.delete(ident, i, axis=0))

    @staticmethod
    def extended(n, i, m):
        ext = Matrix.extend(n,i)
        return Matrix.product(
            ext,
            m,
            ext.inverse()
        )

    @staticmethod
    def vector(v):
        n = len(v)
        va = np.array(v)
        return Matrix(np.reshape(va, (n,1)), np.reshape(va, (1,n)) / np.dot(va, va))

    @staticmethod
    def identity(n):
        ident = np.fromfunction(lambda i,j: np.where(i==j, 1.0, 0.0), (n, n))
        return Matrix(ident, ident)
    @staticmethod
    def translate(d):
        n = len(d)
        column = np.reshape(np.array([*d] + [0.0], dtype=float), (n+1,1))
        forward = np.fromfunction(lambda i,j: np.where(i==j,1.0,0.0) + np.where(np.logical_and(j==n, i<n), column, 0.0), (n+1, n+1))
        backward = np.fromfunction(lambda i,j: np.where(i==j,1.0,0.0) + np.where(np.logical_and(j==n, i<n), -column, 0.0), (n+1, n+1))
        return Matrix(forward, backward)

    @staticmethod
    def project(u):
        n = len(u)
        return (np.reshape(u,(n,1)) @ np.reshape(u,(1,n))) / np.dot(u,u)
    @staticmethod
    def reflect(u_tup):
        pu = Matrix.project(np.array(u_tup))
        return Matrix(2*pu, 2*pu) - Matrix.identity(len(u_tup))
    @staticmethod
    def rotate_xy(n,alpha):
        alpha_rad = alpha * np.pi / 180.0
        c, s = np.cos(alpha_rad), np.sin(alpha_rad)
        ret = np.fromfunction(lambda i, j: np.where(i==j, 1.0, 0.0), (n,n))
        ret[0,0] = c
        ret[1,0] = s
        ret[0,1] = -s
        ret[1,1] = c
        return Matrix(ret, np.moveaxis(ret, 1, 0))
    @staticmethod
    def rotate(u_tup, v_tup):
        u = np.array(u_tup)
        v = np.array(v_tup)
        un = u / np.sqrt(np.dot(u,u))
        vn = v / np.sqrt(np.dot(v,v))
        return Matrix.reflect(un+vn) * Matrix.reflect(vn)

    @staticmethod
    def swap_axes(n, i0, j0):
        def value_at(i,j):
            x = np.where(np.logical_or(i==i0,i==j0), -i + (i0+j0), i)
            return np.where(x==j, 1.0, 0.0)
        forward = np.fromfunction(value_at, (n, n))
        return Matrix(forward, forward)

    @staticmethod
    def scale(factors):
        f = np.array(factors)
        n = len(factors)
        forward = np.fromfunction(lambda i,j: np.where(i==j, f, 0.0), (n,n))
        backward = np.fromfunction(lambda i,j: np.where(i==j, 1/f, 0.0), (n,n))
        return Matrix(forward, backward)

    @staticmethod
    def frustum(l, r, b, t, n, f):
        w, h, d = r-l, t-b, f-n
        x0 = 2.*n/w
        x1 = (r+l)/w
        x2 = 2.*n/h
        x3 = (t+b)/h
        x4 = -(f+n)/d
        x5 = -2.*f*n/d
        forward = np.array([[x0, 0., x1 , 0.],
                            [0., x2, x3,  0.],
                            [0., 0., x4,  x5],
                            [0., 0., -1., 0.]])
        backward = np.array([
            [1/x0, 0.0, 0.0, x1/x0],
            [0.0, 1/x2, 0.0, x3/x2],
            [0.0, 0.0, 0.0, -1.0  ],
            [0.0, 0.0, 1/x5, x4/x5]
        ])
        return Matrix(forward, backward)

    @staticmethod
    def rotate3(alpha, norm):
        return Matrix.rotate((0.0,0.0,1.0,0.0), norm) * Matrix.rotate_xy(4, alpha) * Matrix.rotate(norm, (0.0,0.0,1.0,0.0))

    @staticmethod
    def product(m, *ms) -> "Matrix":
        ret = m
        for mat in ms:
            ret = ret * mat
        return ret

    def __mul__(self, m):
        return Matrix(self.coefs @ m.coefs, m.inv_coefs @ self.inv_coefs)
    def __add__(self, m):
        return Matrix(self.coefs + m.coefs, self.inv_coefs + m.inv_coefs)
    def __sub__(self, m):
        return Matrix(self.coefs - m.coefs, self.inv_coefs - m.inv_coefs)
    def __repr__(self):
        return repr((self.coefs, self.inv_coefs))



class Point:
    def __init__(self, x, y):
        self.x = x
        self.y = y
    def __add__(self, p):
        return Point(self.x + p.x, self.y + p.y)
    def __div__(self, x):
        return Point(self.x / x, self.y / x)
    def __mul__(self, x):
        return Point(self.x * x, self.y * x)
    def __str__(self):
        return str((self.x, self.y))
    def __repr__(self):
        return repr((self.x, self.y))



"""
TODO_DOC
"""

from .__prelude__ import logger
from .Threads import GLOBAL_THREAD_POOL, Runnable
from . import Qt

class ComputedValue(Qt.QObject):
    progressed = Qt.Signal()
    ready = Qt.Signal()

    class ReadyRunner:
        def __init__(self, result, value):
            self._result = result
            self._value = value
        def set_next(self, runner):
            self._result.set_runner(runner(self._value))
        def progress(self):
            return 1.0
        def value(self):
            return self._value
        def start(self):
            self._result.progressed.emit()
        def stop(self):
            pass
        def abort(self):
            pass
    class ComputeRunner(Runnable):
        def __init__(self, result, value, compute):
            super().__init__()
            self.setAutoDelete(True)
            self._result = result
            self._value = value
            self._progress = 0.0
            self._compute = compute
            self._next_runner = lambda value: ComputedValue.ReadyRunner(self._result, value)
            self._stop = False
            self._abort = False

        def set_next(self, runner):
            self._next_runner = runner
            self.stop()
        def progress(self):
            return self._progress
        def value(self):
            return self._value
        def run(self):
            progress_gen = self._compute()
            self._result.progressed.emit()
            try:
                while True:
                    self._progress = next(progress_gen)
                    self._result.progressed.emit()
                    if self._stop or self._abort:
                        value = self._value
                        break
            except StopIteration as e:
                value = e.value
            except RuntimeError as e:
                logger.debug("Runner ended without a result. This must mean the application exited early")
                return
            if not self._abort:
                next_runner = self._next_runner(value)
                self._result.set_runner(next_runner)
            else:
                logger.debug("Thread aborted")
        def start(self):
            GLOBAL_THREAD_POOL.start(self)
        def stop(self):
            self._stop = True
        def abort(self):
            logger.debug("Aborting runner")
            self._abort = True
            self.stop()

    def __init__(self, value):
        super().__init__()
        self._current_runner = self.ReadyRunner(self, value)

    def latest(self):
        return self._current_runner.value()
    def progress(self):
        return self._current_runner.progress()
    def set_runner(self, runner):
        old_runner = self._current_runner
        self._current_runner = runner
        runner.start()
        self.ready.emit()
    def set_value(self, value):
        self._current_runner.set_next(lambda _: self.ReadyRunner(self, value))
    def recompute(self, compute):
        self._current_runner.set_next(lambda value: self.ComputeRunner(self, value, compute))



from .CurveSum         import CurveSum
from .CurveEstimate    import CurveEstimate
from .CurveParam       import CurveParam
from .Curve            import Curve
from .LinearCurve      import LinearCurve
from .ScaledStepCurve  import ScaledStepCurve
from .PeriodicCurve    import PeriodicCurve
from .SigmoidCurve     import SigmoidCurve



from typing import Optional
import pathlib
import numpy as np
import xyzservices.providers

from .__prelude__ import Qt, logger

from .drawing import MapScene, GeoScene, MinimapScene, PointsLayer, SelectedBandLayer, layer_types
from .observable import ObservableList, ObservableStruct, dynamic, SELF
from .SelectedBand import SelectedBand
from .ColorMap import ColorMap
from .Dataset import Dataset, AbstractDataset
from .MapPoint import MapPoint
from .MapProfile import MapProfile

class LayerList(ObservableList):
    def __init__(self, selected_band, points, profiles, colormap):
        super().__init__()
        self._points = points
        self._profiles = profiles
        self._selected_band = selected_band
        self._colormap = colormap

    def import_context(self):
        return {
            'points': self._points,
            'profiles': self._profiles,
            'selected_band': self._selected_band,
            'colormap': self._colormap
        }

def clamp_band(self, band):
    if self.dataset is None:
        return 0
    else:
        return max(0, min(band, self.dataset.count-1))

class MaybeBandNumber(ObservableStruct):
    band_number = dynamic.variable()
    source_band_number = dynamic.external()
    is_enabled = dynamic.variable(False)

    def __init__(self, dynamic_band_number):
        super().__init__()
        self._dynamic_source_band_number = dynamic_band_number
        self.band_number = self.source_band_number
        self.dynamic_attribute("band_number").value_changed.connect(self.__on_band_change)
        self.dynamic_attribute("is_enabled").value_changed.connect(self.__on_enabled_change)

    def __on_band_change(self):
        if self.is_enabled:
            self.source_band_number = self.band_number
    def __on_enabled_change(self):
        if not self.is_enabled:
            self.source_band_number = None
        else:
            self.source_band_number = self.band_number

class WindowState(ObservableStruct):
    class CurrentBand(SelectedBand):
        def __init__(self, win_state: "WindowState"):
            super().__init__(win_state.dynamic_attribute("dataset"), win_state.dynamic_attribute("band_number"), win_state.dynamic_attribute("reference_band_number"))

    dataset                = dynamic.variable()
    band_number            = dynamic.filtered_variable(clamp_band)
    reference_band_number  = dynamic.variable(None)
    tile_provider          = dynamic.external()
    map_scene              = dynamic.readonly()
    maybe_reference_band   = dynamic.readonly()
    geo_lut_dataset        = dynamic.variable()

    def __init__(self):
        super().__init__()
        self.dataset = AbstractDataset()
        self.band_number = 0
        self._current_band = WindowState.CurrentBand(self)

        self.band_colormap = ColorMap("grey", np.array([[0, 0, 0], [255, 255, 255]]), xzero=-1.0, xone=1.0)
        self.points = ObservableList()
        self.profiles = ObservableList()
        band = self.current_band()
        self.layers = LayerList(band, self.points, self.profiles, self.band_colormap)
        self.layers[:] = [
            PointsLayer(self.dynamic_attribute("dataset"), self.points, self.profiles),
            SelectedBandLayer(band, self.band_colormap)
        ]
        self._map_scene = MapScene(band, self.points, self.profiles, self.layers, self.band_colormap)
        self._dynamic_tile_provider = SELF.map_scene.tile_cache.provider[self]
        self.tile_provider = xyzservices.providers['OpenStreetMap']["Mapnik"]
        self.minimap_scene = MinimapScene(band, self.map_scene, self.band_colormap)

        self._maybe_reference_band = MaybeBandNumber(self.dynamic_attribute("reference_band_number"))

    @dynamic.memo(SELF.geo_lut_dataset)
    def geo_scene(self):
        if self.geo_lut_dataset is None:
            return None
        return GeoScene(
            self.current_band(), self.geo_lut_dataset,
            self.map_scene.dynamic_attribute("overlay_layer"),
            self.map_scene.dynamic_attribute("tile_cache"),
        )

    def init_from_dict(self, dct):
        self.dataset = Dataset(dct['dataset'])
        self.band_number = dct['band_number']
        self.reference_band_number = dct.get('reference_band_number', None)
        self.maybe_reference_band.band_number = self.reference_band_number
        self.maybe_reference_band.is_enabled = self.reference_band_number is not None
        self.points[:] = [MapPoint.from_dict(p) for p in dct['points']]
        self.profiles[:] = [MapProfile.from_dict(p) for p in dct['profiles']]
        if 'layers' in dct:
            def make_layer(mime_type, data):
                for ltype in layer_types:
                    if mime_type == ltype.__mime_type__:
                        return ltype.from_dict(data, **self.layers.import_context())
                return None
            self.layers[:] = [make_layer(**dct) for dct in dct['layers']]
        if 'band_colormap' in dct:
            self.band_colormap.init_from_dict(dct['band_colormap'])
        geo_lut_dataset = dct.get('geo_lut_dataset', None)
        if geo_lut_dataset is not None:
            geo_lut_dataset = Dataset(geo_lut_dataset)
        self.geo_lut_dataset = geo_lut_dataset
        self.map_scene.init_from_dict(dct['map_scene'])
    def to_dict(self):
        return {
            'dataset': str(pathlib.Path(self.dataset.file).resolve()),
            'geo_lut_dataset': str(self.geo_lut_dataset.file.resolve()) if self.geo_lut_dataset is not None else None,
            'band_number': self.band_number,
            'reference_band_number': self.reference_band_number,
            'points': [p.to_dict() for p in self.points],
            'profiles': [p.to_dict() for p in self.profiles],
            'map_scene': self.map_scene.to_dict(),
            'band_colormap': self.band_colormap.to_dict(),
            'layers': [{"mime_type": l.__mime_type__, "data": l.to_dict()} for l in self.layers]
        }

    def exit(self):
        self.map_scene.selected_band.texture.destroy()

    def current_band(self):
        return self._current_band



import math
import numpy as np
import rasterio.windows

from .__prelude__ import Qt, ComputedValue, bresenham

from .observable import ObservableStruct, dynamic, Unique, SELF
from .MapPoint import MapPoint, PointCurve

def _line_coefs(x0, y0, x1, y1):
    dx = x1-x0
    dy = y1-y0
    div = y0*dx-x0*dy
    return (dy, -dx, div)
def _eval_line(line, p):
    a,b,c = line
    x,y = p
    return a*x+b*y+c
def _crosses_line(line, p1, p2):
    return _eval_line(line, p1) * _eval_line(line, p2) < 0
def _segments_intersection(p0, p1, p2, p3):
    """
    Return the intersection of two 2D segments, or None if they don't intersect
    """
    l0 = _line_coefs(*p0, *p1)
    l1 = _line_coefs(*p2, *p3)
    p2_c = _eval_line(l0, p2)
    p3_c = _eval_line(l0, p3)
    if p2_c*p3_c >= 0 or not _crosses_line(l1, p0, p1):
        return None
    x2, y2 = p2
    x3, y3 = p3
    x = (x2*p3_c - x3*p2_c) / (p3_c-p2_c)
    y = (y2*p3_c - y3*p2_c) / (p3_c-p2_c)
    return (x,y)

class MapProfile(ObservableStruct):
    name                    = dynamic.variable()
    x0                      = dynamic.variable()
    y0                      = dynamic.variable()
    end_points              = dynamic.variable()
    r                       = dynamic.variable()
    color                   = dynamic.variable()
    show_in_map             = dynamic.variable()
    show_spatial_variance   = dynamic.variable(False)
    show_temporal_variance  = dynamic.variable(False)
    focus_point             = dynamic.variable(0.5)

    def __init__(self, name, r, color, show_in_map, x0, y0, end_points = []):
        super().__init__()
        self.name = name
        self.x0 = int(x0)
        self.y0 = int(y0)
        self.end_points = end_points
        self.r = r
        self.color = color
        self.show_in_map = show_in_map
        self.dynamic_attribute("show_spatial_variance").value_changed.connect(self.__on_show_spatial_variance)
        self.dynamic_attribute("show_temporal_variance").value_changed.connect(self.__on_show_temporal_variance)

    @Qt.Slot()
    def __on_show_spatial_variance(self):
        if self.show_spatial_variance:
            self.show_temporal_variance = False
    @Qt.Slot()
    def __on_show_temporal_variance(self):
        if self.show_temporal_variance:
            self.show_spatial_variance = False

    __mime_type__ = "x-application/insarviz/map-profile"
    @staticmethod
    def from_dict(dct: dict):
        ret = MapProfile(dct["name"], dct["r"], Qt.QColor(dct["color"]), dct["show_in_map"], dct["x0"], dct["y0"], dct["end_points"])
        ret.show_spatial_variance = dct.get("show_spatial_variance", False)
        ret.show_temporal_variance = dct.get("show_temporal_variance", False)
        ret.focus_point = dct.get("focus_point", 0.5)
        return ret

    def to_dict(self):
        return {
            "name": self.name,
            "x0": self.x0,
            "y0": self.y0,
            "end_points": self.end_points,
            "r": self.r,
            "color": self.color.name(),
            "focus_point": self.focus_point,
            "show_in_map": self.show_in_map,
            "show_spatial_variance": self.show_spatial_variance,
            "show_temporal_variance": self.show_temporal_variance,
        }

    @dynamic.memo(SELF.focus_point)
    def focus_center(self):
        lengths = []
        x0, y0 = self.x0, self.y0
        for x1, y1 in self.end_points:
            lengths.append(max(abs(x1-x0), abs(y1-y0)))
            x0, y0 = x1, y1
        total_length = sum(lengths)
        if total_length == 0:
            return self.x0, self.y0
        focus = self.focus_point
        x0, y0 = self.x0, self.y0
        for l, (x1, y1) in zip(lengths, self.end_points):
            proportion = l / total_length
            if focus <= proportion:
                scale = focus/proportion
                dx, dy = (x1-x0) * scale, (y1-y0)*scale
                return x0+dx, y0+dy
            else:
                focus -= proportion
            x0, y0 = x1, y1
        return x0,y0

    def __normal_segment_vector(self, x0, y0, x1, y1):
        dx, dy = x1-x0, y1-y0
        rxy = math.sqrt(dx*dx+dy*dy)
        if rxy <= 0.00001:
            return 0.0,0.0
        return -dy*self.r/rxy, dx*self.r/rxy

    def __end_of_path(self, x0, y0, x1, y1, x2, y2):
        vx0, vy0 = self.__normal_segment_vector(x0, y0, x1, y1)
        vx1, vy1 = self.__normal_segment_vector(x1, y1, x2, y2)

        inter = _segments_intersection((x0+vx0, y0+vy0), (x1+vx0,y1+vy0),
                                       (x1+vx1, y1+vy1), (x2+vx1,y2+vy1))
        if inter is not None:
            return inter
        else:
            return (x1+vx1, y1+vy1)

    def __path_segment_to(self, path, x0, y0, x1, y1, x2, y2):
        vx0, vy0 = self.__normal_segment_vector(x0, y0, x1, y1)
        vx1, vy1 = self.__normal_segment_vector(x1, y1, x2, y2)

        inter = _segments_intersection((x0+vx0, y0+vy0), (x1+vx0,y1+vy0),
                                       (x1+vx1, y1+vy1), (x2+vx1,y2+vy1))
        if inter is not None:
            path.lineTo(*inter)
        else:
            # Pitfall : angles are the opposite of what they should be,
            # because the y axis goes down instead of up.
            path.lineTo(x1+vx0, y1+vy0)
            v0_angle = math.atan2(vy0, vx0) * 180.0 / math.pi
            v1_angle = math.atan2(vy1, vx1) * 180.0 / math.pi
            dv_angle = v0_angle - v1_angle
            while dv_angle < 0.0:
                dv_angle += 360.0
            path.arcTo(x1-self.r, y1-self.r, 2*self.r, 2*self.r, -v0_angle, dv_angle)

    @dynamic.method(SELF.color, SELF.r, SELF.show_in_map, SELF.focus_center)
    def draw(self, painter, opacity):
        if not self.show_in_map:
            return

        path = Qt.QPainterPath()
        path.setFillRule(Qt.Qt.FillRule.WindingFill)

        # Okay, this bears some explaining.
        #
        # To properly handle transparency, we need to draw the profile
        # as a single path rather than multiple overlapping segments.
        #
        # To do so, we draw a series of "path segments" between all
        # triplets of consecutive center points (p0, p1, p2), going
        # around the profile.
        #
        # For example, if our profile consists of only two points p0
        # and p1, we start at p1, draw a path segment (p1, p0, p1),
        # and a second path segment (p0, p1, p0) to close the loop.
        #
        # Could we have just drawn some opaque circles and rectangles
        # on a pixmap, and blitted the result with some transparency ?
        # Sure, but this way does more math and is, therefore, better.
        if len(self.end_points) == 0:
            return
        point_loop = list((*self.end_points, *reversed(self.end_points[:-1]), (self.x0, self.y0)))
        x0, y0, x1, y1, x2, y2 = *point_loop[-3 % len(point_loop)], *point_loop[-2], *point_loop[-1]
        path.moveTo(*self.__end_of_path(x0, y0, x1, y1, x2, y2))
        x0, y0, x1, y1 = x1, y1, x2, y2
        for x2, y2 in point_loop:
            self.__path_segment_to(path, x0, y0, x1, y1, x2, y2)
            x0, y0, x1, y1 = x1, y1, x2, y2

        alpha = int(255 * opacity)
        pen = Qt.QColor(self.color.red(),self.color.green(),self.color.blue(), alpha)
        painter.setPen(pen)
        painter.setBrush(pen)

        painter.drawPath(path)
        painter.setPen(Qt.QColor("black"))
        if self.focus_center is not None:
            painter.drawEllipse(Qt.QPoint(*self.focus_center), self.r, self.r)

class FocusPoint(MapPoint):
    profile       = dynamic.readonly()
    name          = dynamic.external()
    color         = dynamic.external()
    r             = dynamic.external()
    show_in_plot  = dynamic.variable(True)
    show_variance = dynamic.external()
    point_curve   = dynamic.variable()
    focus         = dynamic.external()

    def __init__(self, profile):
        ObservableStruct.__init__(self)
        self._profile = profile
        self._dynamic_name           = SELF.name[profile]
        self._dynamic_color          = SELF.color[profile]
        self._dynamic_r              = SELF.r[profile]
        self._dynamic_show_variance  = SELF.show_spatial_variance[profile]
        self._dynamic_focus          = SELF.focus_point[profile]
        self.point_curve = PointCurve()

    @dynamic.memo(SELF.profile.focus_center)
    def x(self):
        return self.profile.focus_center[0]
    @dynamic.memo(SELF.profile.focus_center)
    def y(self):
        return self.profile.focus_center[1]

class BoundingBox:
    def __init__(self, minX, minY, maxX, maxY):
        self.minX = minX
        self.minY = minY
        self.maxX = maxX
        self.maxY = maxY

    @staticmethod
    def centered_at(x, y, r):
        return BoundingBox(x-r, y-r, x+r, y+r)

    def union(self, bb):
        return BoundingBox(min(self.minX, bb.minX), min(self.minY, bb.minY),
                           max(self.maxX, bb.maxX), max(self.maxY, bb.maxY))
    def values(self):
        return [self.minX, self.minY, self.maxX, self.maxY]

class ProfileData(ObservableStruct):
    smoothing_factor        = dynamic.variable()
    selected_band           = dynamic.readonly()
    reference_data          = dynamic.readonly()
    color                   = dynamic.external()
    profile                 = dynamic.readonly()
    computed_values         = dynamic.readonly()
    show_spatial_variance   = dynamic.external()
    show_temporal_variance  = dynamic.external()
    focus_point             = dynamic.external()

    def __init__(self, selected_band, profile, reference_data):
        super().__init__()
        self._selected_band = selected_band
        self._reference_data = reference_data
        self._profile = profile
        self._dynamic_color = Unique(SELF.profile.color)[self]
        self._dynamic_show_spatial_variance = SELF.profile.show_spatial_variance[self]
        self._dynamic_show_temporal_variance = SELF.profile.show_temporal_variance[self]
        self._dynamic_focus_point = SELF.profile.focus_point[self]
        self._computed_values = ComputedValue((None, None))

        self.__unique_radii = Unique(SELF.profile.r)[self]
        self.__unique_radii.value_changed.connect(lambda: self.computed_values.recompute(self.__get_profile_values))

        self.computed_values.recompute(self.__get_profile_values)
        self.computed_values.ready.connect(self.__on_values_ready)

        self.smoothing_factor = 0

    # Do not reference this in other classes, as it may change unexpectedly
    @dynamic.memo()
    def raw_computed_values(self):
        return self.computed_values.latest()

    @dynamic.memo(
        SELF.reference_data.point_values,
        SELF.selected_band.band_number,
        SELF.selected_band.reference_band_number,
        SELF.raw_computed_values
    )
    def values_and_variance_2d(self):
        values, xs = self.raw_computed_values
        if values is None:
            return None
        ys = values[self.selected_band.band_number, :]

        ref_band = self.selected_band.reference_band_number
        if ref_band is None:
            ref_profile = np.zeros_like(ys[:,0])
        else:
            ref_profile = values[ref_band, :, 0]
        ys_adjusted = ys[:, 0]-ref_profile-self.reference_data.point_values[1][self.selected_band.band_number]
        return xs, ys_adjusted, ys[:, 1]

    @dynamic.memo(SELF.values_and_variance_2d)
    def values_2d(self):
        vals = self.values_and_variance_2d
        if vals is None:
            return None
        xs, ys, _1 = vals
        return xs, ys
    @dynamic.memo(SELF.raw_computed_values)
    def temporal_variance(self):
        values, xs = self.raw_computed_values
        if values is None:
            return None
        return np.nanvar(values[:, :, 0], axis=0)
    @dynamic.memo(SELF.temporal_variance, SELF.values_2d)
    def values_2d_temporal(self):
        vals = self.values_2d
        if vals is None:
            return None
        xs, ys = vals
        return xs, ys, self.temporal_variance

    @dynamic.memo(
        SELF.raw_computed_values,
        SELF.reference_data.point_values,
    )
    def values_3d(self):
        values, ys = self.raw_computed_values
        if values is not None:
            xs, zs_ref = self.reference_data.point_values
            zs = values[:,:,0]-np.reshape(zs_ref, (len(zs_ref), 1))
            return xs, ys, zs
        return np.array([0,1]), np.array([0,1]), np.array([[0,0],[0,0]])

    @Qt.Slot()
    def __on_values_ready(self):
        self.dynamic_attribute("raw_computed_values").invalidate()

    def __get_segment_values(self, x0, y0, x1, y1):
        r = self.profile.r
        minX, minY, maxX, maxY = BoundingBox.centered_at(x0, y0, r).union(BoundingBox.centered_at(x1, y1, r)).values()

        # Create a mask of circular shape
        mask_row = np.reshape(np.arange(2*r+1), (1,2*r+1)) - r
        mask_col = np.reshape(np.arange(2*r+1), (2*r+1,1)) - r
        mask = np.where(mask_row*mask_row + mask_col*mask_col <= r*r, 1.0, 0.0)

        win = rasterio.windows.Window(minX, minY, maxX-minX+1, maxY-minY+1)
        centers = np.array([])
        centers = bresenham.line(x0, y0, x1, y1)
        band_profiles = []
        progress = 0.0
        progress_increment = 1.0 / self.selected_band.dataset.count
        for i in range(self.selected_band.dataset.count):
            band_data = self.selected_band.dataset.read(i+1, window=win)
            band_profile = []

            for (x,y) in centers:
                cx_in_band, cy_in_band = x-minX, y-minY
                data_around_point = band_data[cy_in_band-r:cy_in_band+r+1, cx_in_band-r:cx_in_band+r+1]
                masked_data = mask * data_around_point
                mean = np.nanmean(masked_data)
                var = np.nanvar(masked_data)
                band_profile.append((np.nan if mean == 0.0 else mean, var))
            band_profiles.append(band_profile)
            progress += progress_increment
            yield progress
        return np.array(band_profiles)
    def __get_profile_values(self):
        p = self.profile
        segments = []
        x0, y0 = p.x0, p.y0
        progress_increment = 1.0 / len(self.profile.end_points)
        progress = 0.0
        fmt = "Compute values for profile %s : %%p%%" % self.profile.name
        for x1,y1 in self.profile.end_points:
            new_segment = yield from map(lambda seg_progress: (fmt, progress+progress_increment*seg_progress),
                                         self.__get_segment_values(x0, y0, x1, y1))
            segments.append(new_segment)
            progress += progress_increment
            yield fmt, progress
            x0,y0 = x1, y1
        ret = np.concatenate(segments, axis=1)
        return ret, np.arange(ret.shape[1])



from typing import Any
import pyqtgraph
import numpy as np
import time

from .__prelude__ import (
    logger, Qt, CurveEstimate, linmap, Runnable, GLOBAL_THREAD_POOL
)

class FitCurveItem(pyqtgraph.PlotDataItem):
    class RunnerSignals(Qt.QObject):
        updated_estimate = Qt.Signal()
        finished_estimate = Qt.Signal()
    class Runner(Runnable):
        def __init__(self, estimate):
            super().__init__()
            self.stop = False
            self.estimate = estimate
            self._signals = FitCurveItem.RunnerSignals()
        @property
        def updated_estimate(self):
            return self._signals.updated_estimate
        @property
        def finished_estimate(self):
            return self._signals.finished_estimate

        def abort(self):
            self.stop = True

        def run(self):
            i = 0
            start_time = time.perf_counter()
            try:
                while not self.stop:
                    if not self.estimate.improve():
                        logger.debug("Cannot improve estimate further after %d iterations", i)
                        break
                    current_time = time.perf_counter()
                    if current_time - start_time >= 0.1:
                        self.updated_estimate.emit()
                        start_time = current_time
                    i += 1
                self.updated_estimate.emit()
                self.finished_estimate.emit()
            except RuntimeError as e:
                logger.debug("Error while improving curve. Maybe the application exited early")

    finished_estimate = Qt.Signal(Any)
    def __init__(self, curve):
        super().__init__()
        self._curve = curve
        self._runner = None

    def set_fit_points(self, x, y):
        if self._runner is not None:
            self.stop_fit()

        self.from_x, self.to_x = linmap(x[0], x[-1])
        self.from_y, self.to_y = linmap(np.nanmin(y), np.nanmax(y))
        self._x_mapped = self.from_x(x)
        self._y_mapped = self.from_y(y)

    def set_curve(self, curve):
        self.stop_fit()
        self._curve = curve

    @Qt.Slot(str)
    def _on_estimate_change(self):
        x_inp = np.linspace(-1.0, 1.0, num=256)
        xs = self.to_x(x_inp)
        y_out = self._curve.func(x_inp, *self._estimate.current_guess)
        ys = self.to_y(y_out)
        self.setData(x = xs, y = ys)

    def start_fit(self):
        if self._runner is not None:
            self.stop_fit()
        curve = self._curve
        est = CurveEstimate(curve, self._x_mapped, self._y_mapped)
        self._estimate = est
        self._runner = self.Runner(self._estimate)
        self._runner.updated_estimate.connect(self._on_estimate_change)
        def on_finished():
            self.finished_estimate.emit(est)
        self._runner.finished_estimate.connect(on_finished)
        GLOBAL_THREAD_POOL.start(self._runner)

    def stop_fit(self):
        if self._runner is None:
            return
        self._runner.stop = True
        self._runner = None



from contextlib import ExitStack
import ctypes
import numpy as np
from OpenGL import GL
from shiboken6 import VoidPtr

from . import Qt, Matrix, Point, Bound

class GLUniform:
    def __init__(self, name, value):
        self.__name = name
        self.__value = value
        self.__release = lambda: None
        if isinstance(value, Matrix):
            self.__bind_in = self.__bind_matrix
        elif isinstance(value, int):
            self.__bind_in = lambda program: program.setUniformValue1i(self.__name, self.__value)
        elif isinstance(value, float):
            self.__bind_in = lambda program: program.setUniformValue1f(self.__name, self.__value)
        elif isinstance(value, Qt.QColor):
            self.__bind_in = self.__bind_color
        elif isinstance(value, list):
            self.__bind_in = lambda program: program.setUniformValue(self.__name, *self.__value)
        else:
            self.__bind_in = self.__bind_texture
            self.__release = lambda: self.__value.release()

    def __bind_matrix(self, program):
        w, h = self.__value.shape
        if w == 2:
            cons = Qt.QMatrix2x2
        elif w == 3:
            cons = Qt.QMatrix3x3
        else:
            cons = Qt.QMatrix4x4
        program.setUniformValue(self.__name, cons(self.__value.flatten()))
    def __bind_texture(self, program):
        next_id = program.alloc_id()
        self.__value.bind(next_id)
        program.setUniformValue1i(self.__name, next_id)
    def __bind_color(self, program):
        col = self.__value
        program.setUniformValue(self.__name, col.red()/255.0, col.green()/255.0, col.blue()/255.0, col.alpha()/255.0)

    def bind(self, program):
        return self.__bind_in(program)
    def release(self):
        self.__release()

FLOAT_SIZE = ctypes.sizeof(ctypes.c_float)

def splitTriangle(n, points, i0, i1, i2):
    """Split a triangle in 2^n, by splitting its longest side, and
    creating two smaller triangles, then recursively splitting each of
    those 2^(n-1) times.

    The 'points' input is an array of points. i0, i1 and i2 are
    indices into that array, describing a triangle whose longest side
    joins i0 and i2.

    The functions will store new points into that array destructively,
    and return the indices of all newly-created triangles.
    """
    if n==0:
        return [i0,i1,i2]
    else:
        imid = len(points)
        mid = (points[i0] + points[i2]) * 0.5
        points.append(mid)
        new_triangles_1 = splitTriangle(n-1, points, i1, imid, i0)
        new_triangles_2 = splitTriangle(n-1, points, i2, imid, i1)
        return new_triangles_1 + new_triangles_2

DEFAULT_TEXTURE_TO_MODEL = Matrix.product(
    Matrix.translate((-1,-1)),
    Matrix.scale((2,2,1)),
)

class GLProgram(Qt.QOpenGLShaderProgram):
    class __Mesh:
        def __init__(self, program, mesh_type, attribs, vertex_data, index_data = None):
            self.__program = program
            self.__mesh_type = mesh_type
            self.__vao = Qt.QOpenGLVertexArrayObject()
            self.__vao.create()

            gl = program.gl()
            with Bound(self.__vao):
                num_vertices, record_count = vertex_data.shape
                record_size = record_count * FLOAT_SIZE

                vbuf = Qt.QOpenGLBuffer(Qt.QOpenGLBuffer.Type.VertexBuffer)
                vbuf.create()
                vbuf.setUsagePattern(Qt.QOpenGLBuffer.UsagePattern.StaticDraw)
                vbuf.bind()
                vbuf.allocate(bytes(vertex_data), vertex_data.nbytes)

                offset = 0
                for name, attrib_count in attribs:
                    loc = program.attributeLocation(name)
                    if loc == -1:
                        offset += attrib_count
                        continue
                    gl.glEnableVertexAttribArray(loc)
                    gl.glVertexAttribPointer(loc, attrib_count, GL.GL_FLOAT, GL.GL_FALSE, record_size, VoidPtr(offset * FLOAT_SIZE))
                    offset += attrib_count

                if index_data is not None:
                    indices = Qt.QOpenGLBuffer(Qt.QOpenGLBuffer.Type.IndexBuffer)
                    indices.create()
                    indices.setUsagePattern(Qt.QOpenGLBuffer.UsagePattern.StaticDraw)
                    indices.bind()
                    indices.allocate(bytes(index_data), index_data.nbytes)

                    self.__has_indices = True
                    self.__num_vertices = len(index_data)
                else:
                    self.__has_indices = False
                    self.__num_vertices = num_vertices

        def draw(self):
            gl = self.__program.gl()
            with Bound(self.__vao):
                if self.__has_indices:
                    gl.glDrawElements(self.__mesh_type, self.__num_vertices, GL.GL_UNSIGNED_INT, VoidPtr(0))
                else:
                    gl.glDrawArrays(self.__mesh_type, 0, self.__num_vertices)

    class __Activated:
        class __Uniforms:
            def __init__(self, stack, program):
                self.__program = program
                self.__stack = stack

            def __setattr__(self, attr, val):
                if attr.startswith('_'):
                    super().__setattr__(attr, val)
                else:
                    uni = GLUniform(attr, val)
                    self.__stack.enter_context(Bound(uni, self.__program))

        def __init__(self, stack, program):
            self.__uniforms = self.__Uniforms(stack, program)

        @property
        def uniforms(self):
            return self.__uniforms

    def __init__(self, shaders):
        super().__init__()
        self.__context = Qt.QOpenGLContext.currentContext()
        assert self.__context is not None
        self.__exit_stack = ExitStack()
        self.__current_id = 0

        for shader in shaders:
            shader.compile(self)
        self.link()

    def gl(self):
        return self.__context.functions()

    def alloc_id(self):
        ret = self.__current_id
        self.__current_id += 1
        return ret

    def create_mesh(self, mesh_type, attribs, vertex_data, index_data = None):
        return self.__Mesh(self, mesh_type, attribs, vertex_data, index_data = index_data)

    def create_grid_mesh(self, attrib, w, h):
        h_inc = 1.0/h
        w_inc = 1.0/w

        ln = h_inc * np.arange(h+1)
        ln_x0 = np.stack([np.zeros_like(ln), ln], axis=-1)
        ln_x1 = np.stack([w_inc * np.ones_like(ln), ln], axis=-1)

        strip_line = np.reshape(np.stack([ln_x0, ln_x1], axis=1), (1, 2*(h+1), 2))

        a = np.tile(np.array([[[1,1]], [[1,-1]]]), ((w+1)//2,1,1))[:w,:,:]
        b = np.tile(np.array([[[0,0]], [[0,1]]]), ((w+1)//2,1,1))[:w,:,:]
        base = np.reshape(np.arange(w), (w, 1, 1)) * np.array([[[w_inc, 0.0]]])

        grid = np.reshape(a*strip_line + (b + base), (2*w*(h+1), 2)).astype(np.float32)

        return self.__Mesh(self, GL.GL_TRIANGLE_STRIP, [(attrib, 2)], grid)

    def create_square_mesh(self, mesh_type, split = 0, texture_to_model = DEFAULT_TEXTURE_TO_MODEL):
        vpoints = [Point(0.0,0.0), Point(1.0, 0.0), Point(1.0, 1.0), Point(0.0,1.0)]
        vinds = splitTriangle(split, vpoints, 0, 1 ,2)
        vinds += splitTriangle(split, vpoints, 2, 3, 0)
        vertex_data = np.array([
            [*texture_to_model.transform_point((p.x,p.y)), 0.0, p.x, p.y]
            for p in vpoints
        ]).astype(np.float32)
        return self.__Mesh(self, mesh_type, [("vertex_model_coord", 3), ("vertex_texcoord", 2)],
                           vertex_data, np.array(vinds, dtype="uint32"))

    def __enter__(self):
        stack = self.__exit_stack.__enter__()
        stack.enter_context(Bound(self))
        self.__current_id = 0
        return self.__Activated(stack, self)
    def __exit__(self, *args):
        return self.__exit_stack.__exit__(*args)

def example():
    prog = GLProgram([...])

    mesh = prog.create_mesh(
        GL.GL_TRIANGLES,
        [("vertex_model_coord", 3), ("vertex_tex_coord", 2)],
        np.array([[...], [...], ...]),
        index_data = np.array([0,1,2,...], dtype="uint32")
    )

    with prog as gl:
        gl.uniforms.opacity = 0.5
        gl.uniforms.main_texture = tex
        mesh.draw()



# type: ignore[reportIncompatibleMethodOverride]
from typing import Optional, Any

from .__prelude__ import Qt

from .Schema import Schema

class SchemaItemModel(Qt.QAbstractItemModel):
    def __init__(self, schema: Schema, model: Any):
        super().__init__(None)
        self._schema: Schema = schema
        self._mime_types: list[str] = [typ for typ in schema.supported_mime_types()]
        self._root_node: Schema.Node = schema.make_node(self, model)
        self._root_node.set_links(Qt.QModelIndex(), Qt.QModelIndex())

    def node_at(self, index: Qt.QModelIndex) -> Schema.Node:
        if index.isValid():
            ret = index.internalPointer()
        else:
            ret = self._root_node
        return ret

    def rowCount(self, index: Qt.QModelIndex = Qt.QModelIndex()) -> int:
        node = self.node_at(index)
        return node.rowCount()
    def columnCount(self, index = Qt.QModelIndex()) -> int:
        node = self.node_at(index)
        return node.columnCount()

    def removeRows(self, row: int, count: int, parent: Qt.QModelIndex):
        node = self.node_at(parent)
        return node.removeRows(row, count)

    def data(self, index: Qt.QModelIndex, role: Qt.Qt.ItemDataRole):
        node = self.node_at(index)
        return node.data(role)
    def setData(self, index: Qt.QModelIndex, value: Any, role: Qt.Qt.ItemDataRole):
        node = self.node_at(index)
        return node.setData(value, role)
    def flags(self, index: Qt.QModelIndex) -> Qt.Qt.ItemFlag:
        node = self.node_at(index)
        return node.flags()

    def supportedDropActions(self) -> Qt.Qt.DropAction:
        return Qt.Qt.DropAction.CopyAction | Qt.Qt.DropAction.MoveAction
    def supportedDragActions(self) -> Qt.Qt.DropAction:
        return Qt.Qt.DropAction.CopyAction | Qt.Qt.DropAction.MoveAction

    def mimeTypes(self) -> list[str]:
        return self._mime_types
    def mimeData(self, indices: list[Qt.QModelIndex]) -> Optional[Qt.QMimeData]:
        index_list = [index for index in indices]
        if len(index_list) != 1:
            # No multiple drags yet
            return None
        [index] = index_list
        node = self.node_at(index)
        return node.mime_export()
    def canDropMimeData(self, mimeData: Qt.QMimeData, action: Qt.Qt.DropAction, row: int, column: int, parent: Qt.QModelIndex) -> bool:
        node = self.node_at(parent)
        if row == -1 and column == -1:
            return node.can_mime_import_self(mimeData, action)
        else:
            return node.can_mime_import_child(mimeData, action, row, column)
    def dropMimeData(self, mimeData: Qt.QMimeData, action: Qt.Qt.DropAction, row: int, column: int, parent: Qt.QModelIndex):
        node = self.node_at(parent)
        if row == -1 and column == -1:
            return node.mime_import_self(mimeData, action)
        else:
            return node.mime_import_child(mimeData, action, row, column)

    def parent(self, index: Qt.QModelIndex) -> Qt.QModelIndex:
        node = self.node_at(index)
        return node.parent_index()
    def index(self, row: int, column: int, parent: Qt.QModelIndex = Qt.QModelIndex()) -> Qt.QModelIndex:
        node = self.node_at(parent)
        return node.child(row, column).self_index()

class SchemaItemDelegate(Qt.QStyledItemDelegate):
    needPersistent = Qt.Signal(Qt.QModelIndex)

    def __init__(self, item_model: SchemaItemModel, parent = None):
        super().__init__(parent)
        self._item_model: SchemaItemModel = item_model

    def createEditor(self, parent: Qt.QWidget, option: Qt.QStyleOptionViewItem, index: Qt.QModelIndex) -> Optional[Qt.QWidget]:
        node = self._item_model.node_at(index)
        editor = node.createEditor(parent)
        if editor is not None:
            return editor
        return super().createEditor(parent, option, index)

    def setEditorData(self, editor: Qt.QWidget, index: Qt.QModelIndex) :
        node = self._item_model.node_at(index)
        if not node.setEditorData(editor):
            super().setEditorData(editor, index)
    def setModelData(self, editor: Qt.QWidget, model: Qt.QAbstractItemModel, index: Qt.QModelIndex):
        node = self._item_model.node_at(index)
        if not node.setModelData(editor):
            super().setModelData(editor, model, index)

    def destroyEditor(self, editor: Qt.QWidget, index: Qt.QModelIndex):
        if self._item_model.node_at(index).destroyEditor():
            super().destroyEditor(editor, index)

    def eventFilter(self, editor, event):
        if isinstance(editor, Qt.QDialog):
            if event.type() in (Qt.QEvent.Type.Hide, Qt.QEvent.Type.Destroy):
                return super().eventFilter(editor, event)
            if editor.event(event):
                return True
        return super().eventFilter(editor, event)

    def sizeHint(self, option: Qt.QStyleOptionViewItem, index: Qt.QModelIndex) -> Qt.QSize:
        editor = self._item_model.node_at(index).editor()
        if editor is not None:
            return editor.sizeHint()
        return super().sizeHint(option, index)

    def paint(self, painter: Qt.QPainter, option: Qt.QStyleOptionViewItem, index: Qt.QModelIndex):
        node = self._item_model.node_at(index)
        if node.data(Qt.Qt.ItemDataRole.DisplayRole) is None:
            self.needPersistent.emit(index)
        else:
            if not node.paint(painter, option):
                super().paint(painter, option, index)

class SchemaView(Qt.QTreeView):
    selection_changed = Qt.Signal(Any)

    def __init__(self, schema: Schema, model: Any, **kwargs):
        super().__init__(**kwargs)
        self.setHeaderHidden(True)
        self.setDragDropMode(self.DragDropMode.DragDrop)
        self.setDefaultDropAction(Qt.Qt.DropAction.MoveAction)
        self.setContextMenuPolicy(Qt.Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self._make_custom_context_menu)
        self.showDropIndicator()

        self._item_model: SchemaItemModel = SchemaItemModel(schema, model)
        self.setModel(self._item_model)
        item_delegate = SchemaItemDelegate(self._item_model, self)
        self.setItemDelegate(item_delegate)
        item_delegate.needPersistent.connect(lambda index: self.openPersistentEditor(index))

        self._item_model.rowsAboutToBeRemoved.connect(lambda *__args__: self.selectionModel().clear())
        self.selectionModel().currentChanged.connect(self.__on_current_changed)

    @Qt.Slot(Any, Any)
    def __on_current_changed(self, current, previous):
        current = self._item_model.node_at(current)
        return self.selection_changed.emit(current)

    @Qt.Slot(Qt.QPoint) # type: ignore
    def _make_custom_context_menu(self, position):
        item_index = self.indexAt(position)
        node = self._item_model.node_at(item_index)
        menu = node.context_menu(self)
        if menu is not None:
            menu.exec_(self.mapToGlobal(position))



from .layer              import Layer, SelectedBandLayer, PointsLayer, layer_types, RasterLayer, RasterRGBLayer, SwipeLayer
from .GLDrawing          import GLDrawing
from .Scene              import Scene
from .MapScene           import MapScene
from .MinimapScene       import MinimapScene
from .ProfileScene       import ProfileScene
from .GeoScene           import GeoScene



from typing import Any
import pyqtgraph
import numpy as np
import datetime
from pyqtgraph.GraphicsScene import exportDialog

from .__prelude__       import ( Qt, Container, Leaf, SchemaView, ListSchema, MapPoint, FocusPoint, ItemChooser, IconLabel, icon_with_label, SELF )
from .PointSchema       import point_schema, focus_point_schema
from .FitCurveItem      import FitCurveItem
from .PointDataItem     import PointDataItem, PointData
from .SelectedBandLine  import SelectedBandLine

class TemporalPlotWidget(Qt.QWidget):
    class ChildrenOfInterest:
        def __init__(self):
            self.plot_widget: pyqtgraph.PlotWidget
            self.schema_view: SchemaView
            self.cursor_view: SchemaView
            self.lock_axes_checkbox: Qt.QCheckBox
            self.reference_chooser: ItemChooser
            self.export_button: Qt.QPushButton

    def __init__(self, selected_band, points, profiles):
        super().__init__()
        self.__points = points
        self.__focus_points = profiles.map(FocusPoint)
        self.__selected_band = selected_band
        self.__selected_band.fieldChanged.connect(self.__on_band_change)

        self.__cursor_point = MapPoint("Cursor",0,0,0,Qt.QColor('red'))
        self.__cursor_point.show_in_plot = False
        self.__reference_data = PointData(self.__selected_band, None)

        skel = Container(Qt.QHBoxLayout, (), [
            (Leaf(pyqtgraph.PlotWidget, (), id="plot_widget"), {"stretch": 1}),
            Container(Qt.QVBoxLayout, (), [
                Leaf(Qt.QCheckBox, ("Lock Axes",), id="lock_axes_checkbox"),

                    icon_with_label("insarviz:ref.png", "Reference Point"),
                    Leaf(ItemChooser, (self.__points,), id="reference_chooser"),


                    icon_with_label("insarviz:points.png", "Points"),
                    (Leaf(SchemaView, (ListSchema("points", point_schema), self.__points), id="schema_view"),
                     {"stretch": 1}),

                    icon_with_label("insarviz:profile.png", "Focus Points"),
                    (Leaf(SchemaView, (ListSchema("points", focus_point_schema), self.__focus_points), id="profile_schema_view"),
                     {"stretch": 1}),

                    icon_with_label("insarviz:cursor.png", "Cursor Point"),
                    (Leaf(SchemaView, (point_schema, self.__cursor_point), id="cursor_view"),
                     {"stretch": 0}),
                    (Leaf(Qt.QWidget, ()), {"stretch": 2}),

                    Leaf(Qt.QPushButton, ("Export...",), id="export_button"),
            ])
        ])
        self.widgets = self.ChildrenOfInterest()
        skel.create_in(self, self.widgets)

        self.make_point_item(self.__cursor_point)
        self.__point_items = self.__points.map(self.make_point_item, self.delete_point_item)
        self.__focus_point_items = self.__focus_points.map(self.make_point_item, self.delete_point_item)

        self.__selected_band_line = SelectedBandLine(self.__selected_band.dynamic_attribute("dataset"), self.__selected_band.dynamic_attribute("band_number"))
        self.__selected_band_line.sigDragged.connect(self.__on_band_dragged)
        self.__reference_band_line = SelectedBandLine(self.__selected_band.dynamic_attribute("dataset"), self.__selected_band.dynamic_attribute("reference_band_number"))
        self.__reference_band_line.sigDragged.connect(self.__on_reference_dragged)
        self.__dynamic_units = SELF.dataset.value_units[self.__selected_band]
        self.__dynamic_units.drive(lambda units: self.widgets.plot_widget.setLabel('left', f"LOS displacement ({units})" if units is not None else "No units"))
        self.widgets.plot_widget.addItem(self.__selected_band_line)
        self.widgets.plot_widget.addItem(self.__reference_band_line)
        self.widgets.plot_widget.setMouseEnabled(x=False, y=True)
        self.widgets.plot_widget.setBackground('w')
        self.widgets.plot_widget.setForegroundBrush(Qt.Qt.GlobalColor.transparent)
        self.widgets.reference_chooser.currentIndexChanged.connect(self.__on_reference_selected)
        self.widgets.export_button.clicked.connect(self.__export)
        self.widgets.lock_axes_checkbox.checkStateChanged.connect(self.__on_lock_axes)


    @Qt.Slot()
    def __export(self):
        self.__export_dialog = exportDialog.ExportDialog(self.widgets.plot_widget.plotItem.scene())
        self.__export_dialog.show(self.widgets.plot_widget)

    @Qt.Slot(Any)
    def __on_lock_axes(self, checked):
        if checked == Qt.Qt.CheckState.Checked:
            self.widgets.plot_widget.disableAutoRange()
        else:
            self.widgets.plot_widget.enableAutoRange()

    @Qt.Slot(Any)
    def __on_band_dragged(self, band):
        nearest = self.__selected_band.dataset.nearest_band_to_timestamp(band.getPos()[0])
        self.__selected_band.band_number = int(nearest)
    @Qt.Slot(Any)
    def __on_reference_dragged(self, band):
        nearest = self.__selected_band.dataset.nearest_band_to_timestamp(band.getPos()[0])
        self.__selected_band.reference_band_number = int(nearest)

    @Qt.Slot(str)
    def __on_band_change(self, field, old_value):
        if field == "dataset":
            if self.__selected_band.dataset.has_band_dates:
                self.widgets.plot_widget.setAxisItems({"bottom": pyqtgraph.DateAxisItem(text='Date', units='yyyy-mm-dd')})
        if field == "band_number":
            self.__selected_band_line.setPos(self.__selected_band.timestamp)

    @Qt.Slot(int)
    def __on_reference_selected(self, index):
        ref_index = self.widgets.reference_chooser.itemData(index)
        old_point = self.__reference_data.point
        if old_point is not None:
            old_point.is_reference = False
        if ref_index is None or ref_index == -1:
            self.__reference_data.set_point(None)
        else:
            self.__reference_data.set_point(self.__points[ref_index])
            self.__points[ref_index].is_reference = True

    def set_cursor_coords(self, px):
        if px is None:
            self.__cursor_point.show_in_plot = False
        else:
            self.__cursor_point.x, self.__cursor_point.y = px
            self.__cursor_point.show_in_plot = True

    def make_point_item(self, p):
        ret = PointDataItem(PointData(self.__selected_band, p), self.__reference_data)
        def on_points_hovered(_a, points, _b):
            def point_text(p):
                date = datetime.datetime.fromtimestamp(p.pos().x()).strftime("%Y-%m-%d")
                index = p.index()
                return f"Point at {date} (band {index}) : {ret.point_data.point_ys[0][p.index()]}"
            self.widgets.plot_widget.plotItem.getViewBox().setToolTip(
                "\n".join([point_text(p) for p in points]))
        ret.sigPointsHovered.connect(on_points_hovered)
        ret.visibleChanged.connect(lambda: self.widgets.plot_widget.autoRange())
        for item in ret.items:
            self.widgets.plot_widget.addItem(item)
        return ret

    def delete_point_item(self, i, item):
        for item in item.items:
            self.widgets.plot_widget.removeItem(item)



import pyqtgraph
import datetime

from .__prelude__ import Qt

class SelectedBandLine(pyqtgraph.InfiniteLine):
    def __init__(self, dynamic_dataset, dynamic_band_number):
        super().__init__(angle=90, movable=True)
        self._dynamic_dataset = dynamic_dataset
        self._dynamic_band_number = dynamic_band_number
        self._dynamic_band_number.value_changed.connect(self._set_band_number)
        self._set_band_number()

    @property
    def dataset(self):
        return self._dynamic_dataset.value
    @property
    def band_number(self):
        return self._dynamic_band_number.value

    @Qt.Slot()
    def _set_band_number(self):
        if self.band_number is None:
            self.setVisible(False)
            return
        else:
            self.setVisible(True)
        timestamp = self.dataset.band_timestamps[self.band_number]
        self.setPos(timestamp)
        tooltip = f"Band #{self.band_number+1}"
        if self.dataset.has_band_dates:
            date = datetime.datetime.fromtimestamp(timestamp).strftime("%Y-%m-%d")
            tooltip = f"{tooltip} (at {date})"
        self.setToolTip(tooltip)



from .observable   import ObservableList, ObservableStruct, SELF, EACH
from .SelectedBand import SelectedBand
from .drawing      import Scene, MapScene, GeoScene, MinimapScene, ProfileScene, PointsLayer, SelectedBandLayer, RasterLayer, RasterRGBLayer, SwipeLayer, Layer
from .WindowState  import WindowState, MaybeBandNumber
from .MapPoint     import MapPoint, PointCurve, PointData
from .MapProfile   import MapProfile, ProfileData, FocusPoint
from .ColorMap     import colormaps
from .Dataset      import Dataset
from .curve        import CurveEstimate, CurveParam, CurveSum, LinearCurve, SigmoidCurve, PeriodicCurve, ScaledStepCurve

from . import observable



from typing import TypeVar
from .__prelude__ import SelectedBand, Matrix, Point, Qt, dynamic
from .GLDrawing import GLDrawing, InitializedDrawing

GenericScene = TypeVar("GenericScene", bound="Scene", covariant=True)

class SceneDrawing[GenericScene](InitializedDrawing[GenericScene]):
    def __init__(self, context: Qt.QOpenGLContext, scene: GenericScene):
        super().__init__(context, scene)

    @property
    def world_to_clip(self):
        return Matrix.identity(4)

    def clip_to_model(self, clip_x, clip_y):
        clip_to_model = Matrix.product(self.world_to_clip, self.drawing.model_to_world).inverse()

        v1_model = clip_to_model.transform_point((clip_x, clip_y, -0.5))
        v2_model = clip_to_model.transform_point((clip_x, clip_y, 0.5))
        ray_direction = v2_model - v1_model

        zMul = -v2_model[2]/ray_direction[2]
        return (v2_model[0]+ray_direction[0]*zMul, v2_model[1]+ray_direction[1]*zMul)

class Scene(GLDrawing):
    center          = dynamic.variable(Point(0.0, 0.0))
    heightUnits     = dynamic.variable(0.0)
    distance        = dynamic.variable(0.0)
    yaw             = dynamic.variable(0.0)
    pitch           = dynamic.variable(0.0)
    model_to_world  = dynamic.variable(Matrix.identity(4))

    def startDrag(self,  __dragMode__: Qt.Qt.KeyboardModifier, dragPrev: tuple[float, float]):
        return dragPrev
    def whileDrag(self, __dragMode__: Qt.Qt.KeyboardModifier, dragPrev: tuple[float, float], __dragNext__: tuple[float, float], /):
        return dragPrev
    def endDrag(self, __dragMode__: Qt.Qt.KeyboardModifier, __dragPrev__: tuple[float, float], __dragNext__: tuple[float, float], /):
        return None
    def abortDrag(self,  __dragMode__: Qt.Qt.KeyboardModifier, dragPrev: tuple[float, float]):
        pass

    def initializeDrawing(self, context, /):
        return SceneDrawing(context, self)



import numpy as np

from .__prelude__ import ObservableStruct

TOLERANCE = 1e-6

class CurveEstimate(ObservableStruct):
    def __init__(self, curve, xs, ys):
        super().__init__()
        self.curve = curve
        self._current_params = np.array([param.initial for param in curve.params])
        self._min_params = np.array([param.min_bound for param in curve.params])
        self._max_params = np.array([param.max_bound for param in curve.params])
        not_nans = ~np.isnan(ys)
        self.xs = xs[not_nans]
        self.ys = ys[not_nans]

    def error(self, params = None):
        if params is None:
            params = self._current_params
        return self.curve.func(self.xs, *params) - self.ys
    def objective(self, params = None):
        err = self.error(params)
        return np.sum(err*err)

    @property
    def current_guess(self):
        return self._current_params

    def _clipped_step(self, delta):
        return np.clip(self._current_params + delta,
                       a_min = self._min_params,
                       a_max = self._max_params)

    def _linear_search(self, root_obj, grad):
        eta = 1.0
        best_obj = root_obj
        best_params = self._current_params
        while True:
            guess_params = self._clipped_step(-eta*grad)
            guess_obj = self.objective(guess_params)
            if guess_obj >= best_obj:
                break
            best_obj = guess_obj
            best_params = guess_params
            eta *= 2.0
        if eta == 1.0:
            while eta > 1e-20:
                guess_params = self._clipped_step(-eta*grad)
                guess_obj = self.objective(guess_params)
                if guess_obj < best_obj:
                    best_obj = guess_obj
                    best_params = guess_params
                    break
                eta *= 0.5

        if best_obj < root_obj:
            dparams = self._current_params - best_params
            self._current_params = best_params
            self.fieldChanged.emit('current_guess', None)
            return root_obj - best_obj > TOLERANCE
        else:
            return False

    def _obj_der(self, grad, delta):
        params = self._clipped_step(delta)
        err = self.error(params)
        d_obj = np.sum(2 * self.curve.jacobian(self.xs, *params) * err, axis=1)
        return -np.dot(grad, d_obj)

    def _data_at(self, grad, epsilon):
        delta = -epsilon*grad
        params = self._clipped_step(delta)
        obj = self.objective(params)
        der = self._obj_der(grad, delta)
        return params, obj, der

    def _bisection_search(self, root_obj, grad, right_epsilon):
        left_epsilon = 0.0
        left_params, left_obj, left_der = self._data_at(grad, left_epsilon)

        mid = right_epsilon
        mid_params, mid_obj, mid_der = self._data_at(grad, mid)
        while abs(mid_obj - left_obj) > TOLERANCE:
            if mid_obj > left_obj or mid_der >= 0.0:
                right_epsilon = mid
            else:
                left_epsilon, left_params, left_obj, left_der = mid, mid_params, mid_obj, mid_der
            mid = (right_epsilon + left_epsilon) * 0.5
            mid_params, mid_obj, mid_der = self._data_at(grad, mid)

        self._current_params = left_params
        return abs(root_obj - left_obj) > TOLERANCE

    def improve(self):
        err = self.error()
        d_fi = self.curve.jacobian(self.xs, *self._current_params)
        d_obj = 2 * d_fi * err
        grad = np.sum(d_obj, axis=1)
        norm2_grad = np.dot(grad, grad)
        cross_d2_obj = np.tensordot(grad, d_fi, axes=([0],[0]))
        cross_epsilon = np.sum(cross_d2_obj * cross_d2_obj)
        epsilon = norm2_grad / (2*(cross_epsilon + self.curve.hessian_epsilon(self.xs, err, grad, *self._current_params)))
        root_obj = np.sum(err*err)

        hess_params = self._clipped_step(-epsilon*grad)
        hess_obj = self.objective(hess_params)
        if epsilon > 0.0 and root_obj > hess_obj:
            self._current_params = hess_params
            return root_obj - hess_obj > TOLERANCE
        else:
            # if epsilon > 0.0:
            #     return self._bisection_search(root_obj, grad, epsilon)
            return self._linear_search(root_obj, grad)



from ..__prelude__ import *
from ..ItemChooser import ItemChooser
from ..WidgetTree  import Container, Leaf
from ..ComputedProgressBar import ComputedProgressBar
from ..IconLabel import IconLabel
from ..GLWidget import GLWidget
from ..schema import (
    StructRoles, StructSchema, MaybeStructSchema, FloatSliderSchema, BoolSchema,
    SchemaView, ListSchema
)

def BoldLabel(label):
    ret = Qt.QLabel(label)
    ret.setStyleSheet("font-weight: bold; min-height: 20px;")
    return ret
def icon_with_label(icon, label):
    return Container(Qt.QHBoxLayout, (), [
        Leaf(IconLabel, (Qt.QIcon(icon),)),
        (Leaf(BoldLabel, (label,)), {"stretch":1}),
    ])



from .__prelude__ import (
    Qt, Matrix, destroy_texture,
    dynamic, SELF)
from .Layer import Layer, InitializedLayer

STACK_SHADER = """
#version 330
{geo_color_decl}
uniform sampler2D {temp_texture};
uniform mat3 {geo_to_image};

float {geo_height}(vec3 geo_coords) {{
  float dst_height = 0.0;
  {modify_dst_height}
  return dst_height;
}}

vec4 {geo_color}(vec3 geo_coords, inout float up_opacity) {{
  vec4 dst_color = vec4(0.0);
  vec4 src_color;
  {modify_dst}
  vec3 image_coords = {geo_to_image} * geo_coords;
  src_color = texture({temp_texture}, image_coords.xy / image_coords.z);
  dst_color = mix(dst_color, src_color, src_color.w);
  return dst_color;
}}
"""

class LayerStack(Layer):
    layers = dynamic.readonly()
    temp_point = dynamic.external()
    temp_profile = dynamic.external()
    dataset = dynamic.external()

    def __init__(self, dyn_dataset, dyn_temp_point, dyn_temp_profile, layers):
        super().__init__()
        self._dynamic_temp_point = dyn_temp_point
        self._dynamic_temp_profile = dyn_temp_profile
        self._dynamic_dataset = dyn_dataset
        self._layers = layers
        for layer in self.layers:
            layer.renderChanged.connect(self.renderChanged.emit)

        self.layers.endInsertRange.connect(self.__on_layers_inserted)
        self.layers.beginRemoveRange.connect(self.__on_layers_removed)
        self.layers.endRemoveRange.connect(self.__invalidate_shaders)
        self.layers.beginReplaceRange.connect(self.__before_layers_replaced)
        self.layers.endReplaceRange.connect(self.__after_layers_replaced)

    @Qt.Slot()
    def __invalidate_shaders(self):
        self.dynamic_attribute("shaders").invalidate()
        self.renderChanged.emit()

    @Qt.Slot(int, int)
    def __on_layers_inserted(self, start, length):
        for layer in self.layers[start:start+length]:
            layer.renderChanged.connect(self.renderChanged.emit)
        self.__invalidate_shaders()
    @Qt.Slot(int, int)
    def __on_layers_removed(self, start, end):
        for layer in self.layers[start:end]:
            layer.renderChanged.disconnect(self.renderChanged.emit)
    @Qt.Slot(int, int)
    def __before_layers_replaced(self, start, end):
        for layer in self.layers[start:end]:
            layer.renderChanged.disconnect(self.renderChanged.emit)
    @Qt.Slot(int, int)
    def __after_layers_replaced(self, start, end):
        for layer in self.layers[start:end]:
            layer.renderChanged.connect(self.renderChanged.emit)
        self.__invalidate_shaders()

    @dynamic.memo(SELF.dataset)
    def geo_to_image(self):
        w,h = self.dataset.size
        return Matrix.product(
            self.dataset.pixel_to_crs,
            Matrix.scale((w,h,1))
        ).inverse()

    @dynamic.memo()
    def shaders(self):
        modifications = []
        height_modifications = []
        declarations = []
        for layer in self._layers:
            modifications.append(
                "src_color = {geo_color}(geo_coords, up_opacity); dst_color = vec4(mix(dst_color.xyz, src_color.xyz, src_color.w), max(src_color.w, dst_color.w));".format(
                    geo_color = layer.local_ident("geo_color")
                )
            )
            declarations.append(
                "vec4 {geo_color}(vec3 geo_coords, inout float up_opacity);".format(
                    geo_color = layer.local_ident("geo_color")
                )
            )

            height_modifications.append(
                "dst_height += {geo_height}(geo_coords);".format(
                    geo_height = layer.local_ident("geo_height")
                )
            )
            declarations.append(
                "float {geo_height}(vec3 geo_coords);".format(
                    geo_height = layer.local_ident("geo_height")
                )
            )

        ret = [
            shader
            for layer in self._layers
            for shader in layer.shaders
        ] + [
            STACK_SHADER.format(
                modify_dst = '\n'.join(reversed(modifications)),
                modify_dst_height = '\n'.join(reversed(height_modifications)),
                geo_color_decl = '\n'.join(declarations),
                **self.local_idents("geo_color", "geo_height", "geo_to_image", "temp_texture")
            )
        ]

        return ret

    def GL_initialize(self, context):
        return InitializedLayerStack(context, self)

class InitializedLayerStack(InitializedLayer):
    def __init__(self, context, layer):
        super().__init__(context, layer)
        self.__temp_context = self.GL_create_shared_context()

        self.__initialized_layers = [
            None for layer in self.layer.layers
        ]
        self.__removed_layers = []
        self.layer.layers.endInsertRange.connect(self.__on_insert_layers)
        self.layer.layers.endRemoveRange.connect(self.__on_remove_layers)
        self.layer.layers.endReplaceRange.connect(self.__on_replace_layers)

    @Qt.Slot(int, int)
    def __on_insert_layers(self, start, count):
        self.__initialized_layers[start:start] = [ None for _ in range(count) ]
        self.dynamic_attribute("GL_initialized_layers").invalidate()
    @Qt.Slot(int, int)
    def __on_remove_layers(self, start, end):
        self.__remove_layers(self.__initialized_layers[start:end])
        self.__initialized_layers[start:end] = []
        self.dynamic_attribute("GL_initialized_layers").invalidate()
    @Qt.Slot(int, int)
    def __on_replace_layers(self, start, end):
        self.__remove_layers(self.__initialized_layers[start:end])
        self.__initialized_layers[start:end] = [ None for _ in range(start, end) ]
        self.dynamic_attribute("GL_initialized_layers").invalidate()

    def __GL_initialize_layer(self, i, layer):
        ini_layer = layer.GL_initialize(self.context)
        ini_layer.dynamic_attribute("GL_setup").value_changed.connect(self.__layer_needs_setup)
        self.__initialized_layers[i] = ini_layer
    @Qt.Slot()
    def __layer_needs_setup(self):
        self.dynamic_attribute("GL_setup").invalidate()
    def __remove_layers(self, layers):
        self.__removed_layers += layers
        for layer in layers:
            if layer is not None:
                layer.dynamic_attribute("GL_setup").value_changed.disconnect(self.__layer_needs_setup)

    def __GL_free_removed(self):
        for layer in self.__removed_layers:
            if layer is not None:
                layer.GL_free()
        self.__removed_layers = []

    @dynamic.memo()
    def GL_initialized_layers(self):
        self.__GL_free_removed()
        for i, layer in enumerate(self.layer.layers):
            ini_layer = self.__initialized_layers[i]
            if ini_layer is None:
                self.__GL_initialize_layer(i, layer)
        return self.__initialized_layers

    @dynamic.memo(SELF.layer.dataset, SELF.layer.temp_point, SELF.layer.temp_profile,
                     destroy = destroy_texture)
    def GL_temp_texture(self):
        def do_paint(painter):
            if self.layer.temp_point is not None:
                self.layer.temp_point.draw(painter, 1.0)
            if self.layer.temp_profile is not None:
                self.layer.temp_profile.draw(painter, 1.0)
        return self.GL_paint_into_texture(*self.layer.dataset.size, self.__temp_context, do_paint)

    @dynamic.memo(SELF.GL_initialized_layers, SELF.GL_temp_texture, SELF.layer.geo_to_image)
    def GL_setup(self):
        all_setups = [
            layer.GL_setup for layer in self.GL_initialized_layers
        ]
        tex = self.GL_temp_texture
        def doit(program):
            for setup in all_setups:
                setup(program)
            uni = self.local_uniforms(program.uniforms)
            uni.temp_texture = tex
            uni.geo_to_image = self.layer.geo_to_image
        return doit

    def GL_free(self):
        self.__GL_free_removed()
        self.dynamic_attribute("GL_setup").clear()
        self.dynamic_attribute("GL_temp_texture").clear()
        self.dynamic_attribute("GL_initialized_layers").clear()
        for layer in self.__initialized_layers:
            if layer is not None:
                layer.GL_free()



import json

from PySide6.QtCore import Qt, QMimeData, QSize
from PySide6.QtGui import QColor, QBrush, QPen
from PySide6.QtWidgets import QColorDialog, QDialog
from .Schema import Schema

class ColorSchema(Schema):
    class ColorNode(Schema.Node):
        def __init__(self, item_model, schema, model):
            super().__init__(item_model, schema)
            self.set_model(model)

        def flags(self):
            return Qt.ItemFlags.ItemIsEditable | super().flags()
        def data(self, role):
            if role == Qt.ItemDataRole.DisplayRole:
                return self.model
            if role == Qt.ItemDataRole.EditRole:
                return self.model
            if role == Qt.ItemDataRole.SizeHintRole:
                return QSize(0, 20)
            return None
        def setData(self, value, role):
            if role == Qt.ItemDataRole.EditRole:
                self.requestModelChange.emit(value)
                return True
            return False

        def createEditor(self, parent):
            ret = QColorDialog(parent)
            ret.setModal(True)
            ret.setSizeGripEnabled(True)
            return ret
        def destroyEditor(self):
            return True
        def setEditorData(self, editor):
            editor.setCurrentColor(self.model)
            return True
        def setModelData(self, editor):
            if editor.result() == QDialog.DialogCode.Accepted:
                self.requestModelChange.emit(editor.currentColor())
            return True

        def paint(self, painter, option):
            painter.save()
            brush = QBrush(self.model)
            painter.setBrush(brush)
            painter.setPen(QPen(Qt.GlobalColor.black, 1, Qt.PenStyle.SolidLine))
            painter.drawRoundedRect(option.rect.adjusted(5,2,-6,-2), 7, 7)
            painter.restore()
            return True

        def set_model(self, model):
            super().set_model(model)
            self.dataChanged.emit([])

    def make_node(self, item_model, model):
        return self.ColorNode(item_model,self,model)

    def supported_mime_types(self):
        return set(["application/x-insarviz-color"])
    def can_model_from_mime(self, mimeData):
        return mimeData.hasFormat("application/x-insarviz-color")
    def model_from_mime(self, mimeData):
        data_bytes = mimeData.data("application/x-insarviz-color").data()
        [r,g,b,a] = json.loads(bytearray(data_bytes))
        return QColor(r,g,b,a)
    def can_mime_from_model(self, model):
        return True
    def mime_from_model(self, model):
        mimeData = QMimeData()
        dic = [model.red(), model.green(), model.blue(), model.alpha()]
        mimeData.setData("application/x-insarviz-color", bytes(json.dumps(dic), "utf-8"))
        return mimeData



from .Schema import Schema

class UnionSchema(Schema):
    class UnionSchemaNode(Schema.Node):
        def __init__(self, item_model, schema, model):
            super().__init__(item_model, schema)
            self._alt_node: Schema.Node
            self.set_model(model)

        def set_model(self, model):
            classSchema = self._schema._alts.get(model.__class__)
            if classSchema is None:
                self._alt_node = Schema.Node(self.item_model(), self._schema)
            else:
                self._alt_node = classSchema.make_node(self.item_model(), model)
            super().set_model(model)

        def rowCount(self):
            return self._alt_node.rowCount()
        def columnCount(self):
            return self._alt_node.columnCount()
        def removeRows(self, row, count):
            return self._alt_node.removeRows(row, count)
        def flags(self):
            return self._alt_node.flags()
        def data(self, role):
            return self._alt_node.data(role)
        def setData(self, value, role):
            return self._alt_node.setData(value, role)
        def create_child_node(self, row, column):
            return self._alt_node.create_child_node(row, column)

        def createEditor(self, parent):
            return self._alt_node.createEditor(parent)
        def destroyEditor(self):
            return self._alt_node.destroyEditor()
        def editor(self):
            return self._alt_node.editor()
        def setEditorData(self, editor):
            return self._alt_node.setEditorData(editor)
        def setModelData(self, editor):
            return self._alt_node.setModelData(editor)

        def context_menu(self, parent_widget):
            return self._alt_node.context_menu(parent_widget)
        def paint(self, painter, option):
            return self._alt_node.paint(painter, option)

        def set_links(self, parent_index, self_index):
            super().set_links(parent_index, self_index)
            if self._alt_node is not None:
                self._alt_node.set_links(parent_index, self_index)

    def __init__(self, *alts):
        super().__init__()
        self._alts = {klass: schema for klass, schema in alts}

    def make_node(self, item_model, model):
        return self.UnionSchemaNode(item_model, self, model)

    def supported_mime_types(self):
        ret = set()
        for alt in self._alts.values():
            ret = ret | alt.supported_mime_types()
        return ret
    def can_model_from_mime(self, mimeData, /, **kwargs):
        for alt in self._alts.values():
            if alt.can_model_from_mime(mimeData, **kwargs):
                return True
        return False
    def model_from_mime(self, mimeData, /, **kwargs):
        for alt in self._alts.values():
            if alt.can_model_from_mime(mimeData, **kwargs):
                return alt.model_from_mime(mimeData, **kwargs)
        return None
    def can_mime_from_model(self, model):
        for klass, alt in self._alts.items():
            if isinstance(model, klass) and alt.can_mime_from_model(model):
                return True
        return False
    def mime_from_model(self, model):
        for klass, alt in self._alts.items():
            if isinstance(model, klass):
                return alt.mime_from_model(model)
        return None



from typing import Any, Optional
import numpy as np
import pyqtgraph
import datetime
from pyqtgraph.widgets.ColorMapMenu import buildMenuEntryAction

from .__prelude__ import Qt, Matrix, colormaps, logger
from .GLWidget import GLWidget
from .WidgetTree import Container, Leaf
from .ComputedProgressBar import ComputedProgressBar

class BandSlider(pyqtgraph.PlotWidget):
    LINE_PEN = Qt.QColor('black')
    HOVER_LINE_PEN = pyqtgraph.functions.mkPen('red', width=3)
    SELECTED_LINE_PEN = pyqtgraph.functions.mkPen('blue', width=3)
    REFERENCE_LINE_PEN = pyqtgraph.functions.mkPen('purple', width=3)
    DRAGGED_LINE_PEN = pyqtgraph.functions.mkPen('cyan', width=3)

    def __init__(self, selected_band):
        super().__init__()
        # Autorange makes adding lines very slow, and is useless in the band slider
        self.setBackground('w')
        self.setForegroundBrush(Qt.Qt.GlobalColor.transparent)

        self.plotItem.disableAutoRange()
        self.setMaximumHeight(50)
        self.setMouseEnabled(x=False, y=False)
        self.hideAxis('left')
        self._selected_band = selected_band
        self._selected_band.fieldChanged.connect(self._on_band_change)
        self._selected_band_line = pyqtgraph.InfiniteLine(pen=self.SELECTED_LINE_PEN, pos=self._selected_band.timestamp, movable=True)
        self._selected_band_line.addMarker('<|>')
        self._selected_band_line.sigDragged.connect(self._on_drag_selected_band)
        self._selected_band_line.sigPositionChangeFinished.connect(lambda: self._selected_band_line.setPen(self.SELECTED_LINE_PEN))
        self._reference_band_line = pyqtgraph.InfiniteLine(pen=self.REFERENCE_LINE_PEN)
        self._reference_band_line.addMarker('<|>')
        self.__set_reference_pos()
        self._band_lines = []
        self._set_timestamps()

        self._temp_hovered = None

    def __set_reference_pos(self):
        if self._selected_band.reference_timestamp is None:
            self._reference_band_line.setVisible(False)
        else:
            self._reference_band_line.setPos(self._selected_band.reference_timestamp)
            self._reference_band_line.setVisible(True)

    @Qt.Slot(Any)
    def _on_drag_selected_band(self, band):
        timestamp = band.getPos()[0]
        self._selected_band_line.setPen(self.DRAGGED_LINE_PEN)
        self._selected_band.band_number = int(self._selected_band.dataset.nearest_band_to_timestamp(timestamp))
    @Qt.Slot(str)
    def _on_band_change(self, field):
        if field == "band_number":
            self._selected_band_line.setPos(self._selected_band.timestamp)
        if field == "reference_timestamp":
            self.__set_reference_pos()
        if field == "dataset":
            self._set_timestamps()

    def _set_timestamps(self):
        logger.debug("Setting timestamps")
        xs = self._selected_band.dataset.band_timestamps
        for line in self._band_lines:
            self.plotItem.removeItem(line)
        logger.debug("Removed lines")
        self.plotItem.removeItem(self._selected_band_line)
        self.plotItem.removeItem(self._reference_band_line)
        self._band_lines = []
        for x in xs:
            line = pyqtgraph.InfiniteLine(pos=x, pen=self.LINE_PEN, angle=90)
            self._band_lines.append(line)
        logger.debug("Created new lines")

        for line in self._band_lines:
            self.plotItem.addItem(line)
        logger.debug("Added %d new lines", len(self._band_lines))
        self.plotItem.addItem(self._selected_band_line)
        self.plotItem.addItem(self._reference_band_line)
        self.setXRange(xs[0], xs[-1], padding=0.1)
        if self._selected_band.dataset.has_band_dates:
            self.setAxisItems({'bottom': pyqtgraph.DateAxisItem(units='yyyy-mm-dd') })
        logger.debug("Done setting timestamps")


    def _set_hovered(self, hovered):
        if self._temp_hovered is not None:
            self._temp_hovered.setPen(self.LINE_PEN)
        self._temp_hovered = hovered
        if self._temp_hovered is not None:
            self._temp_hovered.setPen(self.HOVER_LINE_PEN)

    def mouseMoveEvent(self, event):
        super().mouseMoveEvent(event)
        timestamp = self.getViewBox().mapSceneToView(event.pos()).x()
        nearest = self._selected_band.dataset.nearest_band_to_timestamp(timestamp)
        self._set_hovered(self._band_lines[nearest])
    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        timestamp = self.getViewBox().mapSceneToView(event.pos()).x()
        nearest = self._selected_band.dataset.nearest_band_to_timestamp(timestamp)
        self._selected_band.band_number = int(nearest)

    def leaveEvent(self, event):
        self._set_hovered(None)

def make_colormap(name, values, mode):
    formatted_values = np.rint(np.array(values)*255).astype(int)
    ticks = np.linspace(0., 1., len(formatted_values))
    return pyqtgraph.ColorMap(ticks, formatted_values, name=name, mapping=mode)

pg_colormaps = { name: make_colormap(name, values, pyqtgraph.ColorMap.CLIP) for name, values in colormaps.items() }
pg_colormaps.update({ 'cyclic '+name: make_colormap('cyclic '+name, values, pyqtgraph.ColorMap.REPEAT) for name, values in colormaps.items() })

class MainWidget(Qt.QSplitter):
    class ChildrenOfInterest:
        def __init__(self):
            self.main_map: GLWidget
            self.position_label: Qt.QLabel
            self.histogram_widget: pyqtgraph.HistogramLUTWidget
            self.legend_widget: pyqtgraph.GraphicsView
            self.autorange_button: Qt.QPushButton
            self.autorange_all_button: Qt.QPushButton
            self.band_slider: BandSlider
            self.compute_band_progress: ComputedProgressBar
            self.full_histogram_progress: ComputedProgressBar

    # Public interface
    pixel_hovered = Qt.Signal(tuple)

    def __init__(self, window_state):
        super().__init__()
        self._state = window_state

        def fixedHeightWidget(height):
            def ret():
                w = Qt.QWidget()
                w.setFixedHeight(height)
                return w
            return ret
        map_skel = Container(Qt.QVBoxLayout, (), [
            Leaf(BandSlider, (self._state.current_band(),), id="band_slider"),
            (Leaf(GLWidget, (self._state.map_scene,), id = "main_map"), {"stretch": 1}),
            Container(Qt.QHBoxLayout, (), [
                Leaf(Qt.QLabel, ("Position : ",), id = "position_label"),
                (Leaf(Qt.QWidget, ()), {"stretch":1}),
                (Leaf(ComputedProgressBar, (), id = "compute_band_progress"), {"stretch":1}),
            ], widget_class = fixedHeightWidget(40)),
        ])
        hist_skel = Container(Qt.QVBoxLayout, (), [
            Leaf(Qt.QPushButton, ("Autorange current band",), id="autorange_button"),
            Leaf(Qt.QPushButton, ("Autorange all bands",), id="autorange_all_button"),
            (Leaf(pyqtgraph.HistogramLUTWidget, (), id="histogram_widget"), {"stretch": 1}),
            Leaf(pyqtgraph.GraphicsView, (), id="legend_widget"),
            Leaf(ComputedProgressBar, (), id="full_histogram_progress")
        ])
        self.widgets = self.ChildrenOfInterest()
        self.setChildrenCollapsible(False)
        self.addWidget(map_skel.create(self.widgets))
        hist_side_widget = hist_skel.create(self.widgets)
        hist_side_widget.setMaximumWidth(200)
        self.addWidget(hist_side_widget)
        self.setCollapsible(1, True)

        self._autorange_shortcut = Qt.QShortcut(Qt.QKeySequence('Ctrl+A'), self.widgets.main_map)
        self._autorange_shortcut.activated.connect(self._on_autorange)
        self._autorange_all_shortcut = Qt.QShortcut(Qt.QKeySequence('Ctrl+Shift+A'), self.widgets.main_map)
        self._autorange_all_shortcut.activated.connect(self._on_autorange_all)
        self._cursor_position = (0.0, 0.0)

        self.widgets.main_map.positionChanged.connect(self._set_cursor_position)
        self._selected_band = self._state.current_band()
        self._selected_band.fieldChanged.connect(self._on_band_changed)
        self._band_colormap = self._state.band_colormap
        self._band_colormap.fieldChanged.connect(self._on_colormap_changed)
        self._setting_levels = False

        self.widgets.autorange_button.clicked.connect(self._on_autorange)
        self.widgets.autorange_all_button.setEnabled(False)
        self.widgets.autorange_all_button.clicked.connect(self._on_autorange_all)

        self._hist_graph: pyqtgraph.BarGraphItem = pyqtgraph.BarGraphItem(x0=[],x1=[],y0=[],y1=[])
        self._hist_graph.setRotation(90)
        self._hist_graph.setOpts(brush = Qt.QColor(220,20,20,100), pen=None)
        self._full_hist_graph: pyqtgraph.BarGraphItem = pyqtgraph.BarGraphItem(x0=[],x1=[],y0=[],y1=[])
        self._full_hist_graph.setRotation(90)
        self._full_hist_graph.setOpts(brush = Qt.QColor(20,20,220,100), pen=None)

        legend_item = pyqtgraph.LegendItem(size=None)
        legend_item.addItem(self._hist_graph, "current band")
        legend_item.addItem(self._full_hist_graph, "all bands")
        self.widgets.legend_widget.setFixedHeight(legend_item.height())
        self.widgets.legend_widget.setMaximumWidth(200)
        self.widgets.legend_widget.setCentralWidget(legend_item)

        self.widgets.histogram_widget.setBackground('w')
        self.widgets.histogram_widget.setForegroundBrush(Qt.Qt.GlobalColor.transparent)
        self.widgets.histogram_widget.vb.addItem(self._full_hist_graph)
        self.widgets.histogram_widget.vb.addItem(self._hist_graph)

        self._init_colormap_menu()

        self.widgets.histogram_widget.sigLevelsChanged.connect(self._on_levels_changed)
        self.widgets.compute_band_progress.set_computed_value(self._state.current_band().computed_image)

    @property
    def resized_viewport(self):
        return self.widgets.main_map.resized_viewport
    @property
    def world_to_clip(self):
        return self.widgets.main_map.world_to_clip

    # Qt events
    def leaveEvent(self, event):
        self.pixel_hovered.emit(None)
    def closeEvent(self, event):
        self.widgets.main_map.close()
        event.accept()

    # Private methods and properties
    def _init_colormap_menu(self):
        grad = self.widgets.histogram_widget.gradient

        grad.menu = pyqtgraph.ColorMapMenu(
            userList=[
                pg_colormaps[name] for name in colormaps.keys()
            ] + [
                pg_colormaps['cyclic '+name] for name in colormaps.keys()
            ])
        grad.menu.removeAction(grad.menu.actions()[0]) # Remove the default "None" colormap
        grad.menu.sigColorMapTriggered.connect(grad.colorMapMenuClicked)
        grad.menu.sigColorMapTriggered.connect(self._on_colormap_selected)

    @Qt.Slot(str)
    def _on_band_changed(self, field):
        self._set_position_label()
        if field == "dataset":
            self._compute_histograms(set_levels = True)
            self._selected_band.dataset.full_histograms.ready.connect(self._set_full_histograms)
            self.widgets.full_histogram_progress.set_computed_value(self._selected_band.dataset.full_histograms)
        if field == "image":
            self._compute_histograms()

    @Qt.Slot()
    def _set_full_histograms(self):
        hist, bins = self._selected_band.dataset.full_histograms.latest()
        self.widgets.autorange_all_button.setEnabled(True)
        self._full_hist_graph.setOpts(x0 = bins[:-1], x1 = bins[1:], y0 = 0, y1 = hist)

    @Qt.Slot(str)
    def _on_colormap_changed(self, field, old_value):
        colormap = self._band_colormap
        if field == "name" and colormap.name != old_value:
            self.widgets.histogram_widget.gradient.menu.sigColorMapTriggered.emit(pg_colormaps[colormap.name])
        if field in ("xzero", "xone") and getattr(colormap, field) != old_value:
            if not self._setting_levels:
                vb = self.widgets.histogram_widget.vb
                lim_min, lim_max = vb.state['limits']['yLimits']
                amplitude = colormap.xone-colormap.xzero
                lower, upper = colormap.xzero-0.1*amplitude, colormap.xone+0.1*amplitude
                set_range = False
                if lim_min > lower:
                    vb.setLimits(yMin=min(lower, lim_min))
                    set_range = True
                if lim_max < upper:
                    vb.setLimits(yMax=max(upper, lim_max))
                    set_range = True
                if set_range:
                    self.widgets.histogram_widget.setHistogramRange(lower, upper)
            self.widgets.histogram_widget.setLevels(min=colormap.xzero, max=colormap.xone)

    @Qt.Slot(Any) #type: ignore
    def _on_colormap_selected(self, colormap):
        self._state.band_colormap.name = colormap.name.removeprefix('cyclic ')
        self._state.band_colormap.image = colormap.getLookupTable(nPts=1024)
        self._state.band_colormap.is_cyclic = colormap.mapping_mode == pyqtgraph.ColorMap.REPEAT

    @Qt.Slot(Any) #type: ignore
    def _on_levels_changed(self, item):
        xmin, xmax = item.getLevels()
        self._setting_levels = True
        self._state.band_colormap.xzero = xmin
        self._state.band_colormap.xone = xmax
        self._setting_levels = False

    @Qt.Slot()
    def _on_autorange(self):
        self._set_histogram_levels(*self._selected_band.histogram)
    @Qt.Slot()
    def _on_autorange_all(self):
        self._set_histogram_levels(*self._selected_band.dataset.full_histograms.latest())

    def _set_histogram_levels(self, hist, bins, set_levels=True):
        hist_norm = hist/np.sum(hist)
        hist_cum = np.cumsum(hist_norm)
        lower, upper = np.array(bins)[np.searchsorted(hist_cum, [0.02, 0.98])]
        self.widgets.histogram_widget.vb.setLimits(yMin=bins[0], yMax=bins[-1])
        self.widgets.histogram_widget.setHistogramRange(lower,upper, padding=0.1)
        if set_levels:
            self.widgets.histogram_widget.setLevels(min=lower, max=upper)

    @Qt.Slot(float, float)
    def _set_cursor_position(self, x_model, y_model):
        self._cursor_position = (x_model, y_model)
        self._set_position_label()
    @Qt.Slot() # type: ignore
    def _set_position_label(self):
        x_model, y_model = self._cursor_position
        x_img, y_img = Matrix.product(
            self._state.map_scene.selected_band.texture.texture_to_image,
            self._state.map_scene.model_to_texture
        ).transform_point((x_model, y_model))
        x_crs, y_crs = self._state.current_band().image_to_crs.transform_point((x_img, y_img))
        img_width, img_height = 1.0, 1.0
        if self._state.dataset is not None:
            img_width, img_height = self._state.dataset.width, self._state.dataset.height
        x_pixel, y_pixel = int(x_img*img_width), int(y_img*img_height)

        elements = []
        if self._state.dataset.has_band_dates:
            elements.append(datetime.datetime.fromtimestamp(self._selected_band.timestamp).strftime('date: %Y-%m-%d'))

        elements.append(f"x: {x_pixel}, y: {y_pixel}")

        if x_img >= 0.0 and x_img < 1.0 and y_img >= 0 and y_img < 1.0:
            try:
                elements.append(f"val: {self._state.current_band().image[y_pixel, x_pixel, 0]:.03f}")
                self.pixel_hovered.emit((x_pixel, y_pixel))
            except:
                pass
        else:
            self.pixel_hovered.emit(None)
        elements.append(f"lon: {x_crs:.03f}, lat: {y_crs:.03f}")

        self.widgets.position_label.setText(", ".join(elements))

    def _compute_histograms(self, set_levels = False):
        hist, bins = self._selected_band.histogram
        if set_levels:
            self._set_histogram_levels(hist, bins)
        self._hist_graph.setOpts(x0 = bins[:-1], x1 = bins[1:], y0 = 0, y1 = hist)



import numpy as np
from numpy.ma import MaskedArray

from .__prelude__ import Matrix, DEMTexture, inOpenGLContext, ComputedValue, Qt, logger

from .observable import ObservableStruct, dynamic, SELF, Unique

def with_derivs(image_masked):
    h, w = image_masked.shape
    image_dx = (np.float32(w) * 0.5) * (np.roll(image_masked, (0,-1), (0,1)) - np.roll(image_masked, (0,1), (0,1)))
    yield 0.33
    image_dy = (np.float32(h) * 0.5) * (np.roll(image_masked, (-1,0), (0,1)) - np.roll(image_masked, (1,0), (0,1)))
    yield 0.66
    mask = np.where(image_masked.mask, np.float32(0.0), np.float32(1.0))
    return np.reshape(
        np.stack([image_masked.data, image_dx, image_dy, mask], axis=-1),
        (h, w, 4))

def destroy_in_context(texture):
    if texture is not None:
        def run():
            texture.destroy()
        inOpenGLContext(run)

DEFAULT_BAND_IMAGE = (np.reshape(np.array([0,0,0,0]), (1,1,4)), [1], [1])

class SelectedBand(ObservableStruct):
    reference_band_number  = dynamic.external()
    band_number            = dynamic.external()
    dataset                = dynamic.external()

    def __init__(self, dynamic_dataset, dynamic_band_number, dynamic_reference_band_number):
        super().__init__()
        self._dynamic_band_number = dynamic_band_number
        self._dynamic_dataset = dynamic_dataset
        self._dynamic_reference_band_number = dynamic_reference_band_number

        self.__image = ComputedValue(DEFAULT_BAND_IMAGE)
        self.__image.ready.connect(self.__image_ready)
        self.__reference_image = ComputedValue(DEFAULT_BAND_IMAGE)
        self.__reference_image.ready.connect(self.__reference_image_ready)

        self.__image_cache = { }
        self.dynamic_attribute("dataset").value_changed.connect(self.__clear_image_cache)
        self.dynamic_attribute("band_number").value_changed.connect(self.__recompute_image)
        self.dynamic_attribute("reference_band_number").value_changed.connect(self.__recompute_reference_image)

    @dynamic.memo()
    def current_computed_image(self):
        return self.__image.latest()
    @Qt.Slot()
    def __image_ready(self):
        self.dynamic_attribute("current_computed_image").invalidate()
    @dynamic.memo()
    def current_computed_reference_image(self):
        return self.__reference_image.latest()
    @Qt.Slot()
    def __reference_image_ready(self):
        self.dynamic_attribute("current_computed_reference_image").invalidate()

    @property
    def computed_image(self):
        return self.__image

    @dynamic.memo(SELF.image, destroy = destroy_in_context)
    def texture(self) -> DEMTexture:
        def run():
            texture = DEMTexture(self.image)
            texture.create()
            return texture
        return inOpenGLContext(run)
    @dynamic.memo(SELF.reference_image, destroy = destroy_in_context)
    def reference_texture(self) -> DEMTexture:
        def run():
            texture = DEMTexture(self.reference_image)
            texture.create()
            return texture
        return inOpenGLContext(run)

    @dynamic.memo(SELF.current_computed_image)
    def histogram(self):
        _, hist, bins = self.current_computed_image
        return hist, bins

    @dynamic.memo(SELF.dataset, SELF.band_number)
    def timestamp(self):
        return self.dataset.band_timestamps[self.band_number]
    @dynamic.memo(SELF.dataset, SELF.reference_band_number)
    def reference_timestamp(self):
        if self.reference_band_number is None:
            return None
        else:
            return self.dataset.band_timestamps[self.reference_band_number]

    @dynamic.memo(SELF.dataset, SELF.band_number)
    def nodata(self):
        if self.band_number is not None:
            return np.float32(self.dataset.nodatavals[self.band_number])
        return np.nan
    @dynamic.memo(SELF.current_computed_image)
    def image(self):
        return self.current_computed_image[0]
    @dynamic.memo(SELF.current_computed_reference_image)
    def reference_image(self):
        return self.current_computed_reference_image[0]

    @dynamic.memo(SELF.dataset)
    def size(self):
        return (self.dataset.width, self.dataset.height)

    @dynamic.memo(SELF.size)
    def image_to_crs(self):
        w, h = self.size
        return Matrix.product(
            self.dataset.pixel_to_crs,
            Matrix.scale((w, h, 1.0))
        )

    def band_title(self):
        if self.band_number is None:
            return None
        title = self.dataset.file
        if self.band_number is not None:
            title = f"{title} - Band {self.band_number+1} of {self.dataset.count}"
        return title

    def __compute_image_function(self, band):
        def compute_image():
            if band in self.__image_cache:
                pass
            else:
                yield "Loading band #%s : %%p%%" % (band+1), 0.0
                band_img = self.dataset.read(band+1, masked=True)
                yield "Compute band normals : %p%", 0.33
                image = yield from map(
                    lambda img_progress: 0.33+0.33*img_progress,
                    with_derivs(band_img))
                yield "Compute band histogram : %p%", 0.66

                vals = band_img.data[~band_img.mask]
                yield 0.8333
                # Friedman-Diaconis estimator for the number of bins
                hist, bins = np.histogram(vals, bins = 'fd', density=True)
                self.__image_cache[band] = (image, hist, bins)
            yield 1.0
            return self.__image_cache[band]
        return compute_image

    @Qt.Slot()
    def __recompute_image(self):
        self.__image.recompute(self.__compute_image_function(self.band_number))
    @Qt.Slot()
    def __recompute_reference_image(self):
        if self.reference_band_number is None:
            self.__reference_image.set_value(DEFAULT_BAND_IMAGE)
        else:
            self.__reference_image.recompute(self.__compute_image_function(self.reference_band_number))
    @Qt.Slot()
    def __clear_image_cache(self):
        self.__image_cache = {}
        self.__recompute_image()
        self.__recompute_reference_image()



from .__prelude__ import Qt, logger

class ComputedProgressBar(Qt.QProgressBar):
    def __init__(self):
        super().__init__()
        self._computed_value = None
        self._set_value()

    def set_computed_value(self, computed_value):
        if self._computed_value is not None:
            self._computed_value.progressed.disconnect(self._set_value)
        self._computed_value = computed_value
        if self._computed_value is not None:
            self._computed_value.progressed.connect(self._set_value)

    @Qt.Slot()
    def _set_value(self):
        if self._computed_value is not None:
            progress = self._computed_value.progress()
            if isinstance(progress, tuple):
                label, progress = progress
                self.setFormat(label)
        else:
            progress = 1.0
        prog_percent = int(100*progress)
        if prog_percent == 100:
            self.setVisible(False)
        else:
            self.setVisible(True)
            if self.value() != prog_percent:
                self.setValue(prog_percent)



from typing import Any

from .__prelude__ import Qt, logger
from .Dynamic import Dynamic
from .Lens import Lens

class ObservableList[T](Qt.QObject):
    """A basic wrapper over raw Python lists (arrays), that provides
    signals to notify when and how the list's contents change.

    beginRemoveRange(start, stop): emitted before removing the range (start,stop).
    endRemoveRange(start, stop): emitted after removing the range (start,stop)
    beginReplaceRange(start, stop): emitted before replacing elements between 'start' and 'stop'
    endReplaceRange(start, stop): emitted after replacing elements between 'start' and 'stop'
    beginInsertRange(start, length): emitted before inserting 'length' elements at index 'start'
    endInsertRange(start, length): emitted after inserting 'length' elements at index 'start'
    """

    beginRemoveRange = Qt.Signal(int, int)
    endRemoveRange = Qt.Signal(int, int)
    beginReplaceRange = Qt.Signal(int, int)
    endReplaceRange = Qt.Signal(int, int)
    beginInsertRange = Qt.Signal(int, int)
    endInsertRange = Qt.Signal(int, int)

    def __init__(self, *elems: T):
        super().__init__()
        self._values: list[T] = [e for e in elems]

    def __len__(self):
        return len(self._values)
    def __getitem__(self, i):
        return self._values[i]
    def __setitem__(self, i, v):
        if isinstance(i, slice):
            start, stop, _ = i.indices(len(self._values))
            values = v
        else:
            start, stop = i,i+1
            values = [v]
        stop_after = start+len(values)
        inserted = max(0, stop_after-stop)
        replaced = len(values)-inserted
        if replaced > 0:
            self.beginReplaceRange.emit(start, start+replaced)
            self._values[start:start+replaced] = values[:replaced]
            self.endReplaceRange.emit(start, start+replaced)
        if stop > stop_after:
            self.beginRemoveRange.emit(stop_after, stop)
            self._values[stop_after:stop] = []
            self.endRemoveRange.emit(stop_after, stop)
        if inserted > 0:
            self.beginInsertRange.emit(stop, inserted)
            self._values[stop:stop] = values[replaced:]
            self.endInsertRange.emit(stop, inserted)

    def __iter__(self):
        return iter(self._values)
    def __repr__(self):
        return repr(self._values)
    def map(self, func, delete_func = None):
        return MapList(func, self, delete_func)
    def map_lens(self, lens):
        return MapList(lens.__getitem__, self)
    def sequence(self):
        return SequenceList(self)

    def append(self, elem):
        self[len(self):] = [elem]

    def import_context(self) -> dict:
        return {}

class MapList(ObservableList):
    beginRemoveRange = Qt.Signal(int, int)
    endRemoveRange = Qt.Signal(int, int)
    beginReplaceRange = Qt.Signal(int, int)
    endReplaceRange = Qt.Signal(int, int)
    beginInsertRange = Qt.Signal(int, int)
    endInsertRange = Qt.Signal(int, int)

    def __init__(self, func, l, delete_func = None):
        super().__init__(*(func(e) for e in l))
        self.__list = l
        self.__func = func
        self.__delete_func = delete_func
        self.__list.beginRemoveRange.connect(self.__begin_remove_range)
        self.__list.endRemoveRange.connect(self.__end_remove_range)
        self.__list.beginInsertRange.connect(self.__begin_insert_range)
        self.__list.endInsertRange.connect(self.__end_insert_range)
        self.__list.beginReplaceRange.connect(self.__begin_replace_range)
        self.__list.endReplaceRange.connect(self.__end_replace_range)

    @Qt.Slot(int, int)
    def __begin_remove_range(self, start, end):
        self.beginRemoveRange.emit(start,end)
    @Qt.Slot(int, int)
    def __end_remove_range(self, start, end):
        if self.__delete_func is not None:
            for i in range(start, end):
                self.__delete_func(i, self._values[i])
        self._values[start:end] = []
        self.endRemoveRange.emit(start,end)
    @Qt.Slot(int, int)
    def __begin_insert_range(self, start, size):
        self.beginInsertRange.emit(start,size)
    @Qt.Slot(int, int)
    def __end_insert_range(self, start, size):
        self._values[start:start] = [self.__func(e) for e in self.__list[start:start+size]]
        self.endInsertRange.emit(start,size)
    @Qt.Slot(int, int)
    def __begin_replace_range(self, start, end):
        self.beginReplaceRange.emit(start,end)
    @Qt.Slot(int, int)
    def __end_replace_range(self, start, end):
        if self.__delete_func is not None:
            for i in range(start, end):
                self.__delete_func(i, self._values[i])
        self._values[start:end] = [self.__func(e) for e in self.__list[start:end]]
        self.endReplaceRange.emit(start,end)

class SequenceList(Dynamic):
    class __ElemCallback:
        def __init__(self, callback, index):
            self.__callback = callback
            self.index = index
        def __call__(self, *args):
            return self.__callback(self.index, *args)

    def __init__(self, l):
        super().__init__()
        self.__list = l
        self.__callbacks = [self.__ElemCallback(self.__elem_changed,i) for i in range(len(self.__list))]
        for cb in self.__callbacks:
            self.__list[cb.index].value_changed.connect(cb)
        self.__values = [sig.value for sig in l]
        self.__list.beginRemoveRange.connect(self.__begin_remove_range)
        self.__list.endRemoveRange.connect(self.__end_remove_range)
        self.__list.endInsertRange.connect(self.__end_insert_range)
        self.__list.beginReplaceRange.connect(self.__begin_replace_range)
        self.__list.endReplaceRange.connect(self.__end_replace_range)

    @property
    def value(self):
        return self.__values

    def __setitem__(self, i, val):
        self.__list[i].value = val

    @Qt.Slot(int, Any)
    def __elem_changed(self, index, old_val):
        old_values = [*self.__values]
        self.__values[index] = self.__list[index].value
        self.value_changed.emit(old_values)

    @Qt.Slot(int, int)
    def __begin_remove_range(self, start, end):
        for cb in self.__callbacks[start:end]:
            self.__list[cb.index].value_changed.disconnect(cb)
    @Qt.Slot(int, int)
    def __end_remove_range(self, start, end):
        self.__callbacks[start:end] = []
        size = end-start
        for cb in self.__callbacks[start:]:
            cb.index -= size
        old_values = [*self.__values]
        self.__values[start:end] = []
        self.value_changed.emit(old_values)
    @Qt.Slot(int, int)
    def __end_insert_range(self, start, size):
        for cb in self.__callbacks[start:]:
            cb.index += size
        self.__callbacks[start:start] = [self.__ElemCallback(self.__elem_changed, i) for i in range(start, start+size)]
        for i in range(start,start+size):
            cb = self.__callbacks[i]
            self.__list[cb.index].value_changed.connect(cb)
        old_values = [*self.__values]
        self.__values[start:start] = [self.__list[i].value for i in range(start, start+size)]
        self.value_changed.emit(old_values)
    @Qt.Slot(int, int)
    def __begin_replace_range(self, start, end):
        for i in range(start, end):
            self.__list[i].value_changed.disconnect(self.__callbacks[i])
    @Qt.Slot(int, int)
    def __end_replace_range(self, start, end):
        for i in range(start, end):
            self.__list[i].value_changed.connect(self.__callbacks[i])
        old_values = [*self.__values]
        self.__values[start:end] = [self.__list[i].value for i in range(start,end)]
        self.value_changed.emit(old_values)

class EACH(Lens):
    def __init__(self, elem_lens):
        self.__elem_lens = elem_lens
    def __getitem__(self, instance):
        return instance.map_lens(self.__elem_lens).sequence()



from .Schema import Schema
from .__prelude__ import ObservableList, Qt

class ListRoles:
    def __init__(self, display):
        self._display = display

    def data(self, role):
        if role == Qt.Qt.ItemDataRole.DisplayRole:
            return self._display
        return None

    def context_menu(self, parent_widget, node):
        return None

class ListIconRoles(ListRoles):
    def __init__(self, display, icon_name):
        super().__init__(display)
        self._icon_name = icon_name
    def data(self, role):
        if role == Qt.Qt.ItemDataRole.DecorationRole:
            return Qt.QIcon(self._icon_name)
        return super().data(role)

class ListSchema(Schema):
    """A schema for lists of objects.

    TODO_DOC
    """
    class ListNode(Schema.Node):
        def __init__(self, item_model, schema, model):
            super().__init__(item_model, schema)
            self._init_model(model)

        @property
        def elemSchema(self):
            return self._schema.elemSchema
        @property
        def list_roles(self):
            return self._schema.list_roles

        @Qt.Slot(int, int) # type: ignore
        def begin_remove_rows(self, start, stop):
            self.beginRemoveRows.emit(start, stop)
        @Qt.Slot(int, int) # type: ignore
        def end_remove_rows(self, start, stop):
            self.endRemoveRows.emit(start, stop)
        @Qt.Slot(int, int) # type: ignore
        def begin_insert_rows(self, start, l):
            self.beginInsertRows.emit(start, l)
        @Qt.Slot(int, int) # type: ignore
        def end_insert_rows(self, start, l):
            self.endInsertRows.emit(start, l)
        @Qt.Slot(int, int) # type: ignore
        def end_replace_rows(self, start, end):
            for i in range(start, end):
                child = self.cached_child(i,0)
                if child is not None:
                    child.set_model(self.model[i])

        def _exit_model(self):
            if hasattr(self, 'model') and isinstance(self.model, ObservableList):
                self.model.beginRemoveRange.disconnect(self.begin_remove_rows)
                self.model.endRemoveRange.disconnect(self.end_remove_rows)
                self.model.beginInsertRange.disconnect(self.begin_insert_rows)
                self.model.endInsertRange.disconnect(self.end_insert_rows)
                self.model.endReplaceRange.disconnect(self.end_replace_rows)
            super()._exit_model()
        def _init_model(self, model):
            super()._init_model(model)
            if isinstance(self.model, ObservableList):
                self.model.beginRemoveRange.connect(self.begin_remove_rows)
                self.model.endRemoveRange.connect(self.end_remove_rows)
                self.model.beginInsertRange.connect(self.begin_insert_rows)
                self.model.endInsertRange.connect(self.end_insert_rows)
                self.model.endReplaceRange.connect(self.end_replace_rows)

        def rowCount(self):
            return len(self.model)
        def columnCount(self):
            return 1
        def removeRows(self, row, count):
            self.model[row:row+count] = []
            return True
        def flags(self):
            return Qt.Qt.ItemFlag.ItemIsDropEnabled | super().flags()
        def data(self, role):
            return self.list_roles.data(role)
        def create_child_node(self, row, __column__):
            child = self.elemSchema.make_node(self._item_model(), self.model[row])
            def on_child_model_change(child_model):
                row = child._self_index.row()
                self.model[row] = child_model
            child.requestModelChange.connect(on_child_model_change)
            return child

        def context_menu(self, parent_widget):
            return self.list_roles.context_menu(parent_widget, self)

        # Drag and drop
        def can_mime_import_self(self, mimeData, __action__, **kwargs):
            # Allow dropping a child onto the list if it is empty
            if len(self.model) == 0:
                return self.elemSchema.can_model_from_mime(mimeData, **kwargs, **self.model.import_context())
            return False
        def mime_import_self(self, mimeData, __action__, **kwargs):
            new_item = self.elemSchema.model_from_mime(mimeData, **kwargs, **self.model.import_context())
            if new_item is None:
                return False
            self.model[0:0] = [new_item]
            return True

        def can_mime_import_child(self, mimeData, __action__, __row__, __column__, **kwargs):
            return self.elemSchema.can_model_from_mime(mimeData, **kwargs, **self.model.import_context())
        def mime_import_child(self, mimeData, __action__, row, __column__, **kwargs):
            new_item = self.elemSchema.model_from_mime(mimeData, **kwargs, **self.model.import_context())
            if new_item is None:
                return False
            self.model[row:row] = [new_item]
            return True

    def __init__(self, display, elemSchema):
        super().__init__()
        if isinstance(display, str):
            self.list_roles = ListRoles(display)
        else:
            self.list_roles = display
        self.elemSchema = elemSchema
    def make_node(self, item_model, model):
        return self.ListNode(item_model, self, model)

    def supported_mime_types(self):
        return self.elemSchema.supported_mime_types()



import numpy as np

from .Curve import Curve

class CurveSum(Curve):
    def __init__(self, *curves):
        super().__init__()
        self.curves = curves

    @property
    def nparams(self):
        return sum((c.nparams for c in self.curves))
    @property
    def params(self):
        return sum((curve.params for curve in self.curves), [])

    def func(self, xs, *params):
        ret = np.zeros_like(xs)
        for curve in self.curves:
            lparams = curve.nparams
            ret = ret + curve.func(xs, *params[:lparams])
            params = params[lparams:]
        return ret
    def jacobian(self, xs, *params):
        ret = []
        for curve in self.curves:
            lparams = curve.nparams
            ret.append(curve.jacobian(xs, *params[:lparams]))
            params = params[lparams:]
        return np.concatenate(ret)

    def hessian_epsilon(self, xs, err, grad, *params):
        ret = 0.0
        for curve in self.curves:
            lparams = curve.nparams
            ret += curve.hessian_epsilon(xs, err, grad[:lparams], *params[:lparams])
            params = params[lparams:]
            grad = grad[lparams:]
        return ret



from typing import Any
from PySide6.QtCore import Qt, QMimeData
from .Schema import Schema

class StringSchema(Schema):
    class StringNode(Schema.Node):
        def __init__(self, item_model, schema, model):
            super().__init__(item_model, schema)
            self.set_model(model)
        def flags(self):
            ret = Qt.ItemFlag.ItemIsEnabled | super().flags()
            if self._schema.is_editable:
                ret = ret | Qt.ItemFlag.ItemIsEditable | Qt.ItemFlag.ItemIsDragEnabled | Qt.ItemFlag.ItemIsDropEnabled | Qt.ItemFlag.ItemIsSelectable
            return ret

        def data(self, role: Qt.ItemDataRole) -> Any:
            if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.ToolTipRole):
                return self.model
            if role == Qt.ItemDataRole.EditRole and self._schema.is_editable:
                return self.model
            return None
        def setData(self, value: Any, role: Qt.ItemDataRole) -> bool:
            if role == Qt.ItemDataRole.EditRole and self._schema.is_editable:
                self.requestModelChange.emit(value)
                return True
            return False
        def set_model(self, model):
            super().set_model(model)
            self.dataChanged.emit([])

    def __init__(self, is_editable = True):
        super().__init__()
        self._is_editable = is_editable
    @property
    def is_editable(self):
        return self._is_editable

    def make_node(self, item_model, model):
        return self.StringNode(item_model,self,model)

    def supported_mime_types(self):
        return set(["application/x-insarviz-string"])
    def can_model_from_mime(self, mimeData: QMimeData) -> bool:
        return mimeData.hasFormat("application/x-insarviz-string")
    def model_from_mime(self, mimeData: QMimeData):
        data_bytes = bytes(mimeData.data("application/x-insarviz-string").data())
        return data_bytes.decode("utf-8")
    def can_mime_from_model(self, __model__):
        return True
    def mime_from_model(self, model):
        mimeData = QMimeData()
        mimeData.setData("application/x-insarviz-string", bytes(model, "utf-8"))
        return mimeData



import numpy as np

from .CurveParam import CurveParam
from .Curve import Curve

class LinearCurve(Curve):
    def __init__(self):
        super().__init__()
        self.velocity = CurveParam("velocity")
        self.bias = CurveParam("bias")

    __mime_type__ = "application/x-insarviz/LinearCurve"

    @property
    def nparams(self):
        return 2
    @property
    def params(self):
        return [self.velocity, self.bias]

    def func(self, xs, v, b):
        return xs * v + b
    def jacobian(self, xs, v, b):
        return np.array([xs, np.ones_like(xs)])
    def hessian_epsilon(self, xs, err, grad, v, b):
        return 0.0



from .PointSchema        import point_schema
from .FitCurveItem       import FitCurveItem
from .PointDataItem      import PointDataItem
from .SelectedBandLine   import SelectedBandLine
from .TemporalPlotWidget import TemporalPlotWidget



from typing import Any, Optional
from PySide6.QtCore import Qt, Slot
from PySide6.QtWidgets import QMenu

from .__prelude__ import ObservableStruct

from .Schema import Schema

class StructRoles:
    @staticmethod
    def flags():
        return Qt.ItemFlag(0)
    @staticmethod
    def field_roles(__field__: str, /) -> Optional[list[Qt.ItemDataRole]]:
        return None
    @staticmethod
    def get_role(__model__: Any, __role__: Qt.ItemDataRole, /) -> Any:
        return None
    @staticmethod
    def set_role(__model__: Any, __role__: Qt.ItemDataRole, __value__: Any, /) -> bool:
        return False
    @staticmethod
    def context_menu(__parent_widget__, __node__, /) -> Optional[QMenu]:
        return None

class StructSchema(Schema):
    class StructNode(Schema.Node):
        def __init__(self, item_model, schema, model):
            super().__init__(item_model, schema)
            self.fieldRows = {name: i for i, (name, _) in zip(range(len(self.fields)), self.fields)}
            self._init_model(model)

        @property
        def fields(self):
            return self._schema.fields
        @property
        def struct_roles(self):
            return self._schema.struct_roles

        @Slot(str) #type: ignore
        def _on_field_changed(self, field):
            roles = self.struct_roles.field_roles(field)
            if roles is not None:
                self.dataChanged.emit(roles)

            if field not in self.fieldRows:
                return
            row = self.fieldRows[field]
            child = self.cached_child(row, 0)
            if child is None:
                return
            child.set_model(getattr(self.model,field))
        def _exit_model(self):
            if hasattr(self, 'model') and isinstance(self.model, ObservableStruct):
                self.model.fieldChanged.disconnect(self._on_field_changed)
            super()._exit_model()
        def _init_model(self, model):
            super()._init_model(model)
            if isinstance(model, ObservableStruct):
                model.fieldChanged.connect(self._on_field_changed)

        def rowCount(self):
            return len(self.fields)
        def flags(self):
            return self.struct_roles.flags() | super().flags()
        def data(self, role):
            return self.struct_roles.get_role(self.model, role)
        def setData(self, value, role):
            return self.struct_roles.set_role(self.model, role, value)
        def context_menu(self, parent_widget):
            return self.struct_roles.context_menu(parent_widget, self)

        def create_child_node(self, row, __column__):
            fieldName, fieldSchema = self.fields[row]
            child = fieldSchema.make_node(self._item_model(), getattr(self.model, fieldName))
            def on_child_edited(child_model):
                setattr(self.model, fieldName, child_model)
            child.requestModelChange.connect(on_child_edited)
            return child

    def __init__(self, klass, struct_roles, *fields):
        self.klass = klass
        self.struct_roles = struct_roles
        self.fields = fields
    def make_node(self, item_model, model):
        return self.StructNode(item_model, self, model)

    def supported_mime_types(self):
        ret = set()
        if self.klass.__mime_type__ is not None:
            ret.add(self.klass.__mime_type__)
        for _, fieldSchema in self.fields:
            ret = ret | fieldSchema.supported_mime_types()
        return ret

    def can_mime_from_model(self, __model__):
        return self.klass.__mime_type__ is not None
    def mime_from_model(self, model):
        return model.to_mime()
    def can_model_from_mime(self, mimeData, **kwargs):
        return self.klass.can_from_mime(mimeData, **kwargs)
    def model_from_mime(self, mimeData, **kwargs):
        return self.klass.from_mime(mimeData, **kwargs)

def MaybeStructRoles(StructRoles):
    class Ret(StructRoles):
        @staticmethod
        def flags():
            return StructRoles.flags() | Qt.ItemFlag.ItemIsUserCheckable
        @staticmethod
        def field_roles(field):
            if field == "is_enabled":
                return [Qt.ItemDataRole.CheckStateRole]
        @staticmethod
        def get_role(model, role):
            if role == Qt.ItemDataRole.CheckStateRole:
                return Qt.Checked if model.is_enabled else Qt.Unchecked
            return StructRoles.get_role(model, role)
        @staticmethod
        def set_role(model, role, value):
            if role == Qt.ItemDataRole.CheckStateRole:
                model.is_enabled = value == Qt.Checked.value
                return True
            return StructRoles.set_role(model, role, value)
    return Ret

class MaybeStructSchema(StructSchema):
    class MaybeStructNode(StructSchema.StructNode):
        def __init__(self, item_model, schema, model):
            super().__init__(item_model, schema, model)

        @Slot(str, Any) #type: ignore
        def _on_field_changed(self, field, old_val):
            if field == "is_enabled":
                count = super().rowCount()
                if old_val != self.model.is_enabled:
                    if self.model.is_enabled:
                        self.beginInsertRows.emit(0, count)
                        self.endInsertRows.emit(0, count)
                    else:
                        self.beginRemoveRows.emit(0, count)
                        self.endRemoveRows.emit(0, count)
                    self.dataChanged.emit([Qt.ItemDataRole.CheckStateRole])
            else:
                if self.model.is_enabled:
                    return super()._on_field_changed(field)

        def rowCount(self):
            return super().rowCount() if self.model.is_enabled else 0

    def __init__(self, klass, struct_roles, *fields):
        super().__init__(klass, MaybeStructRoles(struct_roles), *fields)
    def make_node(self, item_model, model):
        return self.MaybeStructNode(item_model, self, model)



from .                 import Qt, bresenham
from .Matrix           import Matrix
from .DEMTexture       import DEMTexture, unit_square_to_image
from .Point            import Point
from .Bound            import Bound
from .ComputedValue    import ComputedValue
from .Base10Increments import Base10Increments
from .Threads          import Runnable, GLOBAL_THREAD_POOL
from .GLProgram        import GLProgram

def linmap(A,B):
    """Given an interval [A, B], return a pair of affine functions (f, g)
    such that f(A) = -1 and f(B) = 1, and g(-1) = A and g(1) = B.

    These functions allow for efficient computing of affine mappings
    between [A,B] and [-1,1], both for scalar values and NumPy arrays.
    """
    s = A+B
    d = B-A
    if d==0.0:
        # if we map a null interval, map it to the identity
        d = 2.0
        s = 0.0

    a_from = 2.0 / d
    b_from = - s / d
    def from_lin(x):
        return a_from * x + b_from

    a_to = d / 2.0
    b_to = s / 2.0
    def to_lin(x):
        return a_to * x + b_to

    return (from_lin,to_lin)



"""
TODO_DOC
"""
class Bound:
    def __init__(self, bindable, *args):
        self.bindable = bindable
        self.args = args
    def __enter__(self):
        if self.bindable is not None:
            self.bindable.bind(*self.args)
    def __exit__(self, *__args__):
        if self.bindable is not None:
            self.bindable.release()



from .__prelude__ import logger
from . import Qt

class Runnable(Qt.QRunnable):
    def abort(self):
        pass

class Cancellable(Qt.QRunnable):
    def __init__(self, pool, runnable):
        super().__init__(self)
        self._pool = pool
        self._runnable = runnable

    def __hash__(self):
        return id(self)
    def __eq__(self, other):
        return id(self) == id(other)

    def run(self):
        try:
            return self._runnable.run()
        finally:
            self._pool.remove(self)
    def abort(self):
        self._runnable.abort()

class ThreadPool:
    def __init__(self, global_instance):
        self._global_instance = global_instance
        self._threads = {}

    def start(self, runnable):
        cancellable = Cancellable(self, runnable)
        self._threads[cancellable] = cancellable
        self._global_instance.start(cancellable)

    def remove(self, cancellable):
        del self._threads[cancellable]

    def abort_all(self):
        all_threads = [*self._threads.keys()]
        for cancellable in all_threads:
            cancellable.abort()
        self._global_instance.waitForDone(1000)

GLOBAL_THREAD_POOL = ThreadPool(Qt.QThreadPool.globalInstance())



import numpy as np

from .__prelude__ import ObservableStruct

class Curve(ObservableStruct):
    def init_from_dict(self, dct):
        for param in self.params:
            getattr(self, param.name).init_from_dict(dct[param.name])
    def to_dict(self):
        return {
            param.name: getattr(self, param.name).to_dict()
            for param in self.params
        }

    def nparams(self):
        return 0
    def params(self):
        return []

    def func(self, xs):
        return np.zeros_like(xs)
    def jacobian(self, xs):
        return np.reshape(np.array([]), (0, len(xs)))
    def hessian_epsilon(xs, err, grad):
        return 0.0



from .Dynamic           import Dynamic, PureDynamic, MapDynamic, SwitchDynamic
from .Lens              import SELF, Lens, Unique
from .decorators        import dynamic
from .ObservableStruct  import ObservableStruct
from .ObservableList    import ObservableList, EACH



from .SelectedBandLayer  import SelectedBandLayer
from .PointsLayer        import PointsLayer
from .RasterLayer        import RasterLayer
from .RasterRGBLayer     import RasterRGBLayer
from .LayerStack         import LayerStack
from .SwipeLayer         import SwipeLayer
from .Layer              import Layer

layer_types = [SelectedBandLayer, PointsLayer, RasterLayer, SwipeLayer, RasterRGBLayer]



from .__prelude__ import (
    SELF, dynamic
)

from .Layer import Layer, InitializedLayer

SWIPE_SHADER = """
#version 330

uniform mat3 geo_to_model;
uniform mat4 model_to_world;
uniform mat4 world_to_clip;
uniform float {cutoff};
uniform float {hidden_opacity};

float {geo_height}(vec3 geo_coords) {{
   return 0.0;
}}

vec4 {geo_color}(vec3 geo_coords, inout float up_opacity) {{
  vec3 model_coords = geo_to_model * geo_coords;
  vec4 clip_coords = world_to_clip * model_to_world * vec4(model_coords.xy, 0.0, model_coords.z);
  if({hidden_opacity} <= 0.5)
     up_opacity = clip_coords.x / clip_coords.w > {cutoff} ? 1.0 : 0.0;
  return vec4(0.0);
}}
"""

class SwipeLayer(Layer):
    cutoff = dynamic.variable()
    shaders = dynamic.readonly()

    def __init__(self):
        super().__init__()
        self._shaders = [
            SWIPE_SHADER.format(
                **self.local_idents("geo_color", "geo_height", "cutoff", "hidden_opacity")
            )
        ]
        self.cutoff = 0.0
        self.dynamic_attribute("_redraw").value_changed.connect(lambda: self._redraw())

    @dynamic.method(SELF.cutoff, SELF.is_enabled)
    def _redraw(self):
        self.renderChanged.emit()

    def GL_initialize(self, context):
        return InitializedSwipeLayer(context, self)

    __mime_type__ = "x-application/insarviz/SwipeLayer"
    @staticmethod
    def from_dict(dct, **__kwargs__):
        ret = SwipeLayer()
        ret.cutoff = dct['cutoff']
        return ret

    def to_dict(self):
        return {
            "cutoff": self.cutoff
        }


class InitializedSwipeLayer(InitializedLayer):
    def __init__(self, context, layer):
        super().__init__(context, layer)

    @dynamic.memo(SELF.layer.geo_to_clip, SELF.layer.cutoff, SELF.layer.is_enabled)
    def GL_setup(self):
        def doit(program):
            uni = self.local_uniforms(program.uniforms)
            uni.cutoff = 2*self.layer.cutoff-1
            uni.hidden_opacity = 0.0 if self.layer.is_enabled else 1.0
        return doit



"""
In this module, we introduce the notion of "dynamic values".

A dynamic value is a value that changes across time, that can be
listened to, read, and sometimes written to.

In this form, dynamic values may be composed to provide more complex
behaviours, such as switching between value sources, or transforming
values on the fly.
"""

from typing import Any
from collections.abc import Callable
from .__prelude__ import Qt
import weakref

class Dynamic[A](Qt.QObject):
    value_changed = Qt.Signal(Any)

    @property
    def value(self) -> A:
        raise NotImplementedError
    @value.setter
    def value(self, v: A):
        raise NotImplementedError

    def drive(self, func):
        self.value_changed.connect(lambda: func(self.value))
        func(self.value)

    def map(self, func, *other_sigs):
        return MapDynamic(func, self, *other_sigs)
    def join(self):
        return SwitchDynamic(self)
    def then(self, func, *other_sigs):
        return self.map(func, *other_sigs).join()

class MapDynamic(Dynamic):
    def __init__(self, f, *sigs):
        super().__init__()
        self.__f = f
        self.__sigs = sigs
        self.__current_value = self.__f(*(sig.value for sig in self.__sigs))
        for sig in self.__sigs:
            sig.value_changed.connect(self.__on_base_changed)

    @property
    def value(self):
        return self.__current_value

    def __on_base_changed(self, _old):
        old = self.__current_value
        self.__current_value = self.__f(*(sig.value for sig in self.__sigs))
        self.value_changed.emit(old)

class PureDynamic[A](Dynamic[A]):
    def __init__(self, initial):
        super().__init__()
        self.__value: A = initial

    @property
    def value(self):
        return self.__value
    @value.setter
    def value(self, v):
        old = self.__value
        self.__value = v
        self.value_changed.emit(old)

    def dynamic_attribute(self, name):
        return self

class SwitchDynamic[A](Dynamic[A]):
    def __init__(self, sig_sig: Dynamic[Dynamic[A]]):
        super().__init__()
        self.__sig_sig: Dynamic[Dynamic[A]] = sig_sig
        self.__current_sig: Dynamic[A] = sig_sig.value
        if self.__current_sig is None:
            self.__current_sig = PureDynamic(None)
        self.__current_sig.value_changed.connect(self.__on_current_changed)
        self.__sig_sig.value_changed.connect(self.__on_signal_changed)

    @property
    def value(self):
        return self.__current_sig.value
    @value.setter
    def value(self, v):
        self.__current_sig.value = v

    def __on_current_changed(self, old: A):
        self.value_changed.emit(old)
    def __on_signal_changed(self, old_sig: Dynamic[A]):
        self.__current_sig.value_changed.disconnect(self.__on_current_changed)
        new_sig = self.__sig_sig.value
        if new_sig is None:
            self.__current_sig = PureDynamic(None)
        else:
            self.__current_sig = new_sig
        self.__current_sig.value_changed.connect(self.__on_current_changed)
        if old_sig is None:
            self.value_changed.emit(None)
        else:
            self.value_changed.emit(old_sig.value)



from typing import Any, TypeVar
import numpy as np
import math
from OpenGL import GL
import datetime

from .__prelude__ import (
    Matrix, Point, Qt, Base10Increments, Shader, Bound, OpenGLPainter,
    GLProgram, dynamic, SELF,
    destroy_texture, linmap
)
from .Scene import Scene, SceneDrawing

def vec_angle(v1, v2):
    x1, y1 = v1
    x2, y2 = v2
    num = x1*y2-y1*x2
    denom = x1*x2+y1*y2
    return math.atan2(num, denom)

class ProfileScene(Scene):
    profile_data_changed = Qt.Signal()
    pan = dynamic.variable(Point(0.0,0.0))
    profile_data = dynamic.variable()

    def __init__(self, profile_data):
        super().__init__()
        self.profile_data = profile_data
        self.distance = 1.0

        self._line_shaders = [
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Vertex,"profile/line/vertex.glsl"),
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Fragment,"profile/line/fragment.glsl")
        ]
        self._profile_shaders = [
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Vertex,"profile/image/vertex.glsl"),
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Fragment,"profile/image/fragment.glsl"),
            Shader(Qt.QOpenGLShader.ShaderTypeBit.Fragment,"hsv_rgb.glsl"),
        ]

        self.dynamic_attribute('_redraw').value_changed.connect(lambda: self._redraw())
        self.dynamic_attribute('image_to_grid').value_changed.connect(self.__on_coords_changed)

    @Qt.Slot()
    def __on_coords_changed(self):
        grid_center_x, grid_center_y, _ = self.image_to_grid.transform_point((0.5, 0.5, 0.0))
        self.pan = Point(-grid_center_x, -grid_center_y)
        self.distance = 1.0

    @dynamic.method(SELF.model_to_world, SELF.distance, SELF.pan, SELF.profile_data.values_3d, SELF.diff_scale, SELF.profile_data.color, SELF.selected_band_position, SELF.profile_data.focus_point)
    def _redraw(self):
        self.renderChanged.emit()

    @dynamic.memo(SELF.profile_data.selected_band.dataset.has_band_dates)
    def show_time(self):
        if self.profile_data is None:
            def ret(x):
                return f"{x:.6g}"
        elif self.profile_data.selected_band.dataset.has_band_dates:
            def ret(timestamp):
                return datetime.datetime.fromtimestamp(timestamp).strftime("%Y-%m-%d")
        else:
            def ret(x):
                return f"{x:.6g}"
        return ret

    @dynamic.memo(SELF.profile_data.smoothing_factor)
    def diff_scale(self):
        return 0.002*(self.profile_data.smoothing_factor+1)

    @dynamic.memo(SELF.profile_data.values_3d, SELF.profile_data.selected_band.timestamp)
    def selected_band_position(self):
        xs, _, _ = self.profile_data.values_3d
        start,end = np.nanmin(xs), np.nanmax(xs)
        from_x, _ = linmap(start, end)
        return 0.5*(1+from_x(self.profile_data.selected_band.timestamp)), 1/len(xs)

    @dynamic.memo(SELF.pitch, SELF.yaw)
    def model_to_world(self):
        return Matrix.product(
            Matrix.translate((0.0,0.0,8.0)),
            Matrix.rotate3(self.pitch, (-1.0,0.0,0.0,0.0)),
            Matrix.rotate3(self.yaw, (0.0,0.0,1.0,0.0)),
        )
    @dynamic.memo(SELF.distance, SELF.pan, SELF.image_to_grid_scale)
    def grid_to_model(self):
        return Matrix.product(
            Matrix.scale((1./self.distance, 1./self.distance, 1.0, 1.0)),
            Matrix.scale((2,2,1,1)),
            self.image_to_grid_scale.inverse(),
            Matrix.translate((self.pan.x, self.pan.y, 0.0)),
        )

    @dynamic.memo(SELF.profile_data.values_3d)
    def image_to_grid_scale(self):
        if self.profile_data is None:
            return Matrix.identity(4)
        xs, ys, zs = self.profile_data.values_3d
        w, h, d = xs[-1] - xs[0], ys[-1] - ys[0], max(abs(np.nanmin(zs)), abs(np.nanmax(zs)))
        if d < 1e-10:
            d = 1.0
        return Matrix.scale((h, w, d, 1))

    @dynamic.memo(SELF.image_to_grid_scale, SELF.height_scale)
    def image_to_grid(self):
        if self.profile_data is None:
            return Matrix.identity(4)
        xs, ys, zs = self.profile_data.values_3d
        return Matrix.product(
            Matrix.translate((ys[0], xs[0], 0.0)),
            self.image_to_grid_scale,
        )

    @dynamic.memo(SELF.profile_data.values_3d)
    def height_scale(self):
        if self.profile_data is None:
            return 1.0
        _, _, zs = self.profile_data.values_3d
        zscale = max(abs(np.nanmax(zs)), abs(np.nanmin(zs)))
        if zscale == 0.0:
            return 1.0
        return 1/zscale

    def startDrag(self, mode, dragStart):
        return dragStart
    def whileDrag(self, mode, dragStart, dragEnd):
        model_to_grid = Matrix.extend(4,2).inverse() * self.grid_to_model.inverse() * Matrix.extend(4,2)
        xa,ya = model_to_grid.transform_point(dragStart)
        xb,yb = model_to_grid.transform_point(dragEnd)
        self.pan += Point(xb-xa,yb-ya)
        return dragEnd

    def initializeDrawing(self, context):
        return ProfileSceneDrawing(context, self)

class ProfileSceneDrawing(SceneDrawing[ProfileScene]):
    def __init__(self, context, scene):
        super().__init__(context, scene)
        self._lines_program = GLProgram(self.drawing._line_shaders)
        self._profile_program = GLProgram(self.drawing._profile_shaders)

    @dynamic.memo(SELF.drawing.profile_data.values_3d)
    def GL_image_mesh(self):
        if self.drawing.profile_data is None:
            return None
        xs,ys,zs = self.drawing.profile_data.values_3d
        w,h = zs.shape
        return self._profile_program.create_grid_mesh("vertex_texture_coords", w, h)

    @dynamic.memo(SELF.drawing.profile_data.values_3d,
                     destroy = destroy_texture)
    def GL_profile_texture(self):
        if self.drawing.profile_data is None:
            return None

        profile_texture = Qt.QOpenGLTexture(Qt.QOpenGLTexture.Target.Target2D)
        profile_texture.setMagnificationFilter(Qt.QOpenGLTexture.Filter.Linear)
        profile_texture.setWrapMode(Qt.QOpenGLTexture.WrapMode.ClampToEdge)

        xs, ys, zs = self.drawing.profile_data.values_3d
        zs_cum = np.nancumsum(zs, axis=1)
        zs_cum = np.nancumsum(zs_cum, axis=0)
        zs_cumnans = np.cumsum(np.where(np.isnan(zs), 0.0, 1.0), axis=1)
        zs_cumnans = np.cumsum(zs_cumnans, axis=0)
        h,w = zs.shape
        profile_texture.setSize(w,h)
        profile_texture.setFormat(Qt.QOpenGLTexture.TextureFormat.RG32F)
        profile_texture.allocateStorage(Qt.QOpenGLTexture.PixelFormat.RG, Qt.QOpenGLTexture.PixelType.Float32)
        profile_texture.setData(Qt.QOpenGLTexture.PixelFormat.RG, Qt.QOpenGLTexture.PixelType.Float32,
                                np.stack([zs_cum, zs_cumnans], axis=-1).astype(np.float32))
        profile_texture.create()
        return profile_texture

    @dynamic.memo(SELF.viewport_size)
    def world_to_clip(self):
        w,h = self.viewport_size
        hfactor = max(1.0, w/h)/4.0
        vfactor = max(1.0, h/w)/4.0
        near = 4.0*0.005
        return Matrix.frustum(-hfactor*near, hfactor*near,
                              -vfactor*near, vfactor*near,
                              -near, -1000.0)
    @dynamic.memo(SELF.viewport_size)
    def clip_to_pixel(self):
        w,h = self.viewport_size
        return Matrix.scale((w/2.0, -h/2.0, 1.0)) * Matrix.translate((1.0,-1.0))

    def __draw_grid(self, gl):
        model_to_grid = self.drawing.grid_to_model.inverse()
        model_to_clip = Matrix.product(
            self.world_to_clip,
            self.drawing.model_to_world,
        )
        grid_to_clip = Matrix.product(
            model_to_clip,
            self.drawing.grid_to_model
        )

        lines = []
        major_ticks = {}
        for bottomleft, topright in [
                ((-1,-1,-1), (1,1,-1)),
                ((1,1,-1),   (1,-1,1)),
                ((1,-1,1),   (-1,-1,-1)),
                ((-1,1,1),   (1,-1,1)),
                ((-1,1,1),   (-1,-1,-1)),
                ((1,1,-1),   (-1,1,1)),
        ]:
            vertices = [None, None, None]
            for axis in range(3):
                if bottomleft[axis] == topright[axis]:
                    constant_axis = axis
                vertices[axis] = [*bottomleft]
                vertices[axis][axis] = topright[axis]
                vertices[axis] = model_to_clip.transform_point(vertices[axis])[:2]
            if vec_angle(vertices[(constant_axis+1)%3] - vertices[constant_axis],
                         vertices[(constant_axis+2)%3] - vertices[constant_axis]) >= 0:
                continue

            grid_bottomleft = model_to_grid.transform_point(bottomleft)
            grid_topright = model_to_grid.transform_point(topright)

            for axis in range(3):
                if bottomleft[axis] != topright[axis]:
                    line_bottomleft = [*grid_bottomleft, 1.0]
                    line_topright = [*grid_topright, 1.0]
                    line_topright[axis] = line_bottomleft[axis]
                    lines.append(line_bottomleft)
                    lines.append(line_topright)

                    line_bottomleft = [*grid_bottomleft, 1.0]
                    line_topright = [*grid_topright, 1.0]
                    line_bottomleft[axis] = line_topright[axis]
                    lines.append(line_bottomleft)
                    lines.append(line_topright)

                    for inc_type, inc in Base10Increments(grid_bottomleft[axis], grid_topright[axis]):
                        line_weight = 1.0 if inc_type == Base10Increments.MAJOR_INCREMENT else 0.0
                        line_start = [*grid_bottomleft, line_weight]
                        line_end = [*grid_topright, line_weight]
                        line_start[axis] = inc
                        line_end[axis] = inc
                        lines.append(line_start)
                        lines.append(line_end)
                        if inc_type == Base10Increments.MAJOR_INCREMENT:
                            start_ind = tuple(line_start[:3])
                            end_ind = tuple(line_end[:3])
                            ln = [start - 0.1*(end-start) for end,start in zip(end_ind, start_ind)]
                            if start_ind in major_ticks:
                                del major_ticks[start_ind]
                            else:
                                major_ticks[start_ind] = (ln, inc, axis)

                            ln = [end - 0.1*(start-end) for end,start in zip(end_ind, start_ind)]
                            if end_ind in major_ticks:
                                del major_ticks[end_ind]
                            else:
                                major_ticks[end_ind] = (ln, inc, axis)

        line_mesh = self._lines_program.create_mesh(GL.GL_LINES, [("grid_coords", 4)], np.array(lines, dtype=np.float32))
        with self._lines_program as p:
            p.uniforms.grid_to_clip = grid_to_clip
            line_mesh.draw()

        return [ln for ln in major_ticks.values()], grid_to_clip

    def __draw_major_ticks(self, gl, major_ticks, grid_to_clip):
        pixel_ratio = self.context.screen().devicePixelRatio()
        with OpenGLPainter(*[x*pixel_ratio for x in self.viewport_size]) as painter:
            painter.scale(1.0, -1.0)
            painter.setPen(Qt.QColor(0,0,0))
            grid_to_pixel = Matrix.product(
                Matrix.scale((self.viewport_size[0]*pixel_ratio, -self.viewport_size[1]*pixel_ratio, 1.0, 2.0)),
                Matrix.translate((1,1,1)),
                grid_to_clip
            )
            for grid_coord, inc, axis in major_ticks:
                pixel_x, pixel_y, _ = grid_to_pixel.transform_point(grid_coord)
                wHuge, hHuge = 32000, 32000
                rect = Qt.QRectF(pixel_x - wHuge, pixel_y - hHuge, 2*wHuge, 2*hHuge)
                if axis == 1:
                    text = self.drawing.show_time(inc)
                else:
                    text = "{:.6g}".format(inc)
                painter.drawText(rect, text, Qt.QTextOption(Qt.Qt.AlignCenter))
            painter.scale(1.0, -1.0)

    def __draw_image(self, gl):
        tex = self.GL_profile_texture
        mesh = self.GL_image_mesh
        if tex is None or mesh is None:
            return
        with self._profile_program as p:
            image_to_model = Matrix.product(
                self.drawing.grid_to_model,
                self.drawing.image_to_grid,
            )
            image_to_world = Matrix.product(
                self.drawing.model_to_world,
                image_to_model,
            )
            p.uniforms.tex = tex
            p.uniforms.height_scale = self.drawing.height_scale
            p.uniforms.image_to_model = image_to_model
            p.uniforms.model_to_image = image_to_model.inverse()
            p.uniforms.image_to_world = image_to_world
            p.uniforms.world_to_clip = self.world_to_clip
            p.uniforms.diff_scale = self.drawing.diff_scale
            p.uniforms.profile_color = self.drawing.profile_data.color
            band_position, band_width = self.drawing.selected_band_position
            p.uniforms.selected_band_position = band_position
            p.uniforms.selected_band_width = band_width

            p.uniforms.focus_position = self.drawing.profile_data.focus_point
            p.uniforms.focus_width = 1/len(self.drawing.profile_data.values_3d[1])
            mesh.draw()

    def paintGL(self):
        gl = self.context.functions()
        gl.glEnable(GL.GL_DEPTH_TEST)
        gl.glEnable(GL.GL_BLEND)
        gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA)

        gl.glClearColor(1.0,1.0,1.0,1.0)
        gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)

        major_ticks, grid_to_clip = self.__draw_grid(gl)
        self.__draw_major_ticks(gl, major_ticks, grid_to_clip)
        gl.glEnable(GL.GL_DEPTH_TEST)
        gl.glEnable(GL.GL_BLEND)
        gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA)
        self.__draw_image(gl)

    def freeGL(self):
        self.dynamic_attribute("GL_profile_texture").clear()
